from amltk.profiling.memory import Memory
from amltk.profiling.profiler import Profile, Profiler
from amltk.profiling.timing import Timer

__all__ = [
    "Memory",
    "Timer",
    "Profile",
    "Profiler",
]



from __future__ import annotations

from amltk.pipeline.components import (
    Choice,
    Component,
    Fixed,
    Join,
    Searchable,
    Sequential,
    Split,
    as_node,
)
from amltk.pipeline.node import Node, request
from amltk.pipeline.ops import factorize

__all__ = [
    "Node",
    "Component",
    "Split",
    "Choice",
    "Searchable",
    "Sequential",
    "Fixed",
    "Join",
    "request",
    "as_node",
    "factorize",
]



"""A pipeline consists of [`Node`][amltk.pipeline.node.Node]s, which hold
the various attributes required to build a pipeline, such as the
[`.item`][amltk.pipeline.node.Node.item], its [`.space`][amltk.pipeline.node.Node.space],
its [`.config`][amltk.pipeline.node.Node.config] and so on.

The [`Node`][amltk.pipeline.node.Node]s are connected to each in a parent-child
relation ship where the children are simply the [`.nodes`][amltk.pipeline.node.Node.nodes]
that the parent leads to.

To give these attributes and relations meaning, there are various subclasses
of [`Node`][amltk.pipeline.node.Node] which give different syntactic meanings
when you want to construct something like a
[`search_space()`][amltk.pipeline.node.Node.search_space] or
[`build()`][amltk.pipeline.node.Node.build] some concrete object out of the
pipeline.

For example, a [`Sequential`][amltk.pipeline.Sequential] node
gives the meaning that each of its children in
[`.nodes`][amltk.pipeline.node.Node.nodes] should follow one another while
something like a [`Choice`][amltk.pipeline.Choice]
gives the meaning that only one of its children should be chosen.

You will likely never have to create a [`Node`][amltk.pipeline.node.Node]
directly, but instead use the various components to create the pipeline.

??? note "Hashing"

    When hashing a node, i.e. to put it in a `set` or as a key in a `dict`,
    only the name of the node and the hash of its children is used.
    This means that two nodes with the same connectivity will be equalling hashed,

??? note "Equality"

    When considering equality, this will be done by comparing all the fields
    of the node. This include even the `parent` and `branches` fields. This
    means two nodes are considered equal if they look the same and they are
    connected in to nodes that also look the same.
"""  # noqa: E501
# ruff: noqa: PLR0913
from __future__ import annotations

import inspect
import warnings
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from dataclasses import dataclass, field
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Concatenate,
    Generic,
    Literal,
    NamedTuple,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    overload,
)
from typing_extensions import override

from more_itertools import first_true
from sklearn.pipeline import Pipeline as SklearnPipeline

from amltk._functional import classname, funcname, mapping_select, prefix_keys
from amltk._richutil import RichRenderable
from amltk.exceptions import RequestNotMetError
from amltk.optimization.history import History
from amltk.optimization.optimizer import Optimizer
from amltk.scheduling import Task
from amltk.scheduling.plugins import Plugin
from amltk.store import PathBucket
from amltk.types import Config, Item, Seed, Space

if TYPE_CHECKING:
    from typing_extensions import Self

    from ConfigSpace import ConfigurationSpace
    from rich.console import RenderableType
    from rich.panel import Panel

    from amltk.evalutors.evaluation_protocol import EvaluationProtocol
    from amltk.optimization.metric import Metric
    from amltk.optimization.trial import Trial
    from amltk.pipeline.components import Choice, Join, Sequential
    from amltk.pipeline.parsers.optuna import OptunaSearchSpace
    from amltk.scheduling import Scheduler

    NodeLike: TypeAlias = (
        set["Node" | "NodeLike"]
        | tuple["Node" | "NodeLike", ...]
        | list["Node" | "NodeLike"]
        | Callable[..., Item]
        | Item
    )

    SklearnPipelineT = TypeVar("SklearnPipelineT", bound=SklearnPipeline)


class OnBeginCallbackSignature(Protocol):
    """A  calllback to further define control flow from
    [`pipeline.optimize()`][amltk.pipeline.node.Node.optimize].

    In one of these callbacks, you can register to specific `@events` of the
    [`Scheduler`][amltk.scheduling.Scheduler] or [`Task`][amltk.scheduling.Task].

    ```python
    pipeline = ...

    # The callback will get the task, scheduler and the history in which results
    # will be stored
    def my_callback(task: Task[..., Trial.Report], scheduler: Scheduler, history: History) -> None:

        # You can do early stopping based on a target metric
        @task.on_result
        def stop_if_target_reached(_: Future, report: Trial.Report) -> None:
            score = report.metrics["accuracy"]
            if score >= 0.95:
                scheduler.stop(stop_msg="Target reached!"))

        # You could also perform early stopping based on iterations
        n = 0
        last_score = 0.0

        @task.on_result
        def stop_if_no_improvement_for_n_runs(_: Future, report: Trial.Report) -> None:
            score = report.metrics["accuracy"]
            if score > last_score:
                n = 0
                last_score = score
            elif n >= 5:
                scheduler.stop()
            else:
                n += 1

        # Really whatever you'd like
        @task.on_result
        def print_if_choice_made(_: Future, report: Trial.Report) -> None:
            if report.config["estimator:__choice__"] == "random_forest":
                print("yay")

        # Every callback will be called here in the main process so it's
        # best not to do anything too heavy here.
        # However you can also submit new tasks or jobs to the scheduler too
        @task.on_result(every=30)  # Do a cleanup sweep every 30 trials
        def keep_on_ten_best_models_on_disk(_: Future, report: Trial.Report) -> None:
            sorted_reports = history.sortby("accuracy")
            reports_to_cleanup = sorted_reports[10:]
            scheduler.submit(some_cleanup_function, reporteds_to_cleanup)

    history = pipeline.optimize(
        ...,
        on_begin=my_callback,
    )
    ```
    """  # noqa: E501

    def __call__(
        self,
        task: Task[[Trial, Node], Trial.Report],
        scheduler: Scheduler,
        history: History,
    ) -> None:
        """Signature for the callback.

        Args:
            task: The task that will be run
            scheduler: The scheduler that will be running the optimization
            history: The history that will be used to collect the results
        """
        ...


T = TypeVar("T")
ParserOutput = TypeVar("ParserOutput")
BuilderOutput = TypeVar("BuilderOutput")
P = ParamSpec("P")


_NotSet = object()


class RichOptions(NamedTuple):
    """Options for rich printing."""

    panel_color: str = "default"
    node_orientation: Literal["horizontal", "vertical"] = "horizontal"


@dataclass(frozen=True)
class ParamRequest(Generic[T]):
    """A parameter request for a node. This is most useful for things like seeds."""

    key: str
    """The key to request under."""

    default: T | object = _NotSet
    """The default value to use if the key is not found.

    If left as `_NotSet` (default) then an error will be raised if the
    parameter is not found during configuration with
    [`configure()`][amltk.pipeline.node.Node.configure].
    """

    @property
    def has_default(self) -> bool:
        """Whether this request has a default value."""
        return self.default is not _NotSet


def request(key: str, default: T | object = _NotSet) -> ParamRequest[T]:
    """Create a new parameter request.

    Args:
        key: The key to request under.
        default: The default value to use if the key is not found.
            If left as `_NotSet` (default) then the key will be removed from the
            config once [`configure`][amltk.pipeline.Node.configure] is called and
            nothing has been provided.
    """
    return ParamRequest(key=key, default=default)


@dataclass(init=False, frozen=True, eq=True)
class Node(RichRenderable, Generic[Item, Space]):
    """The core node class for the pipeline.

    These are simple objects that are named and linked together to form
    a chain. They are then wrapped in a `Pipeline` object to provide
    a convenient interface for interacting with the chain.
    """

    name: str = field(hash=True)
    """Name of the node"""

    item: Callable[..., Item] | Item | None = field(hash=False)
    """The item attached to this node"""

    nodes: tuple[Node, ...] = field(hash=True)
    """The nodes that this node leads to."""

    config: Config | None = field(hash=False)
    """The configuration for this node"""

    space: Space | None = field(hash=False)
    """The search space for this node"""

    fidelities: Mapping[str, Any] | None = field(hash=False)
    """The fidelities for this node"""

    config_transform: Callable[[Config, Any], Config] | None = field(hash=False)
    """A function that transforms the configuration of this node"""

    meta: Mapping[str, Any] | None = field(hash=False)
    """Any meta information about this node"""

    _NODES_INIT: ClassVar[Literal["args", "kwargs"] | None] = "args"
    """Whether __init__ takes nodes as positional args, kwargs or does not accept it."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(
        panel_color="default",
        node_orientation="horizontal",
    )
    """How to display this node in rich."""

    def __init__(
        self,
        *nodes: Node,
        name: str,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """Initialize a choice.

        Args:
            nodes: The nodes that this node leads to
            name: The name of the node
            item: The item attached to this node
            config: The configuration for this node
            space: The search space for this node
            fidelities: The fidelities for this node
            config_transform: A function that transforms the configuration of this node
            meta: Any meta information about this node
        """
        super().__init__()
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "item", item)
        object.__setattr__(self, "config", config)
        object.__setattr__(self, "space", space)
        object.__setattr__(self, "fidelities", fidelities)
        object.__setattr__(self, "config_transform", config_transform)
        object.__setattr__(self, "meta", meta)
        object.__setattr__(self, "nodes", nodes)

    def __getitem__(self, key: str) -> Node:
        """Get the first from [`.nodes`][amltk.pipeline.node.Node.nodes] with `key`."""
        found = first_true(
            self.nodes,
            None,
            lambda node: node.name == key,
        )
        if found is None:
            raise KeyError(
                f"Could not find node with name {key} in '{self.name}'."
                f" Available nodes are: {', '.join(node.name for node in self.nodes)}",
            )

        return found

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.name == other.name
            and self.item == other.item
            and self.config == other.config
            and self.space == other.space
            and self.fidelities == other.fidelities
            and self.config_transform == other.config_transform
            and self.meta == other.meta
            and self.nodes == other.nodes
        )

    def __or__(self, other: Node | NodeLike) -> Choice:
        from amltk.pipeline.components import as_node

        return as_node({self, as_node(other)})

    def __and__(self, other: Node | NodeLike) -> Join:
        from amltk.pipeline.components import as_node

        return as_node((self, other))  # type: ignore

    def __rshift__(self, other: Node | NodeLike) -> Sequential:
        from amltk.pipeline.components import as_node

        return as_node([self, other])  # type: ignore

    def configure(
        self,
        config: Config,
        *,
        prefixed_name: bool | None = None,
        transform_context: Any | None = None,
        params: Mapping[str, Any] | None = None,
    ) -> Self:
        """Configure this node and anything following it with the given config.

        Args:
            config: The configuration to apply
            prefixed_name: Whether items in the config are prefixed by the names
                of the nodes.
                * If `None`, the default, then `prefixed_name` will be assumed to
                    be `True` if this node has a next node or if the config has
                    keys that begin with this nodes name.
                * If `True`, then the config will be searched for items prefixed
                    by the name of the node (and subsequent chained nodes).
                * If `False`, then the config will be searched for items without
                    the prefix, i.e. the config keys are exactly those matching
                    this nodes search space.
            transform_context: Any context to give to `config_transform=` of individual
                nodes.
            params: The params to match any requests when configuring this node.
                These will match against any ParamRequests in the config and will
                be used to fill in any missing values.

        Returns:
            The configured node
        """
        # Get the config for this node
        match prefixed_name:
            case True:
                config = mapping_select(config, f"{self.name}:")
            case False:
                pass
            case None if any(k.startswith(f"{self.name}:") for k in config):
                config = mapping_select(config, f"{self.name}:")
            case None:
                pass

        _kwargs: dict[str, Any] = {}

        # Configure all the branches if exists
        if len(self.nodes) > 0:
            nodes = tuple(
                node.configure(
                    config,
                    prefixed_name=True,
                    transform_context=transform_context,
                    params=params,
                )
                for node in self.nodes
            )
            _kwargs["nodes"] = nodes

        this_config = {
            hp: v
            for hp, v in config.items()
            if (
                ":" not in hp
                and not any(hp.startswith(f"{node.name}") for node in self.nodes)
            )
        }
        if self.config is not None:
            this_config = {**self.config, **this_config}

        this_config = dict(self._fufill_param_requests(this_config, params=params))

        if self.config_transform is not None:
            this_config = dict(self.config_transform(this_config, transform_context))

        if len(this_config) > 0:
            _kwargs["config"] = dict(this_config)

        return self.mutate(**_kwargs)

    def fidelity_space(self) -> dict[str, Any]:
        """Get the fidelities for this node and any connected nodes."""
        fids = {}
        for node in self.nodes:
            fids.update(prefix_keys(node.fidelity_space(), f"{self.name}:"))

        return fids

    def linearized_fidelity(self, value: float) -> dict[str, int | float | Any]:
        """Get the liniearized fidelities for this node and any connected nodes.

        Args:
            value: The value to linearize. Must be between [0, 1]

        Return:
            dictionary from key to it's linearized fidelity.
        """
        assert 1.0 <= value <= 100.0, f"{value=} not in [1.0, 100.0]"  # noqa: PLR2004
        d = {}
        for node in self.nodes:
            node_fids = prefix_keys(
                node.linearized_fidelity(value),
                f"{self.name}:",
            )
            d.update(node_fids)

        if self.fidelities is None:
            return d

        for f_name, f_range in self.fidelities.items():
            match f_range:
                case (int() | float(), int() | float()):
                    low, high = f_range
                    fid = low + (high - low) * value
                    fid = low + (high - low) * (value - 1) / 100
                    fid = fid if isinstance(low, float) else round(fid)
                    d[f_name] = fid
                case _:
                    raise ValueError(
                        f"Invalid fidelities to linearize {f_range} for {f_name}"
                        f" in {self}. Only supports ranges of the form (low, high)",
                    )

        return prefix_keys(d, f"{self.name}:")

    def iter(self) -> Iterator[Node]:
        """Iterate the the nodes, including this node.

        Yields:
            The nodes connected to this node
        """
        yield self
        for node in self.nodes:
            yield from node.iter()

    def mutate(self, **kwargs: Any) -> Self:
        """Mutate the node with the given keyword arguments.

        Args:
            **kwargs: The keyword arguments to mutate

        Returns:
            Self
                The mutated node
        """
        _args = ()
        _kwargs = {**self.__dict__, **kwargs}

        # If there's nodes in kwargs, we have to check if it's
        # a positional or keyword argument and handle accordingly.
        if (nodes := _kwargs.pop("nodes", None)) is not None:
            match self._NODES_INIT:
                case "args":
                    _args = nodes
                case "kwargs":
                    _kwargs["nodes"] = nodes
                case None if len(nodes) == 0:
                    pass  # Just ignore it, it's popped out
                case None:
                    raise ValueError(
                        "Cannot mutate nodes when __init__ does not accept nodes",
                    )

        # If there's a config in kwargs, we have to check if it's actually got values
        config = _kwargs.pop("config", None)
        if config is not None and len(config) > 0:
            _kwargs["config"] = config

        # Lastly, we remove anything that can't be passed to kwargs of the
        # subclasses __init__
        _available_kwargs = inspect.signature(self.__init__).parameters.keys()  # type: ignore
        for k in list(_kwargs.keys()):
            if k not in _available_kwargs:
                _kwargs.pop(k)

        return self.__class__(*_args, **_kwargs)

    def copy(self) -> Self:
        """Copy this node, removing all links in the process."""
        return self.mutate()

    def path_to(self, key: str | Node | Callable[[Node], bool]) -> list[Node] | None:
        """Find a path to the given node.

        Args:
            key: The key to search for or a function that returns True if the node
                is the desired node

        Returns:
            The path to the node if found, else None
        """
        # We found our target, just return now

        match key:
            case Node():
                pred = lambda node: node == key
            case str():
                pred = lambda node: node.name == key
            case _:
                pred = key

        for path, node in self.walk():
            if pred(node):
                return path

        return None

    def walk(
        self,
        path: Sequence[Node] | None = None,
    ) -> Iterator[tuple[list[Node], Node]]:
        """Walk the nodes in this chain.

        Args:
            path: The current path to this node

        Yields:
            The parents of the node and the node itself
        """
        path = list(path) if path is not None else []
        yield path, self

        for node in self.nodes:
            yield from node.walk(path=[*path, self])

    @overload
    def find(self, key: str | Node | Callable[[Node], bool], default: T) -> Node | T:
        ...

    @overload
    def find(self, key: str | Node | Callable[[Node], bool]) -> Node | None:
        ...

    def find(
        self,
        key: str | Node | Callable[[Node], bool],
        default: T | None = None,
    ) -> Node | T | None:
        """Find a node in that's nested deeper from this node.

        Args:
            key: The key to search for or a function that returns True if the node
                is the desired node
            default: The value to return if the node is not found. Defaults to None

        Returns:
            The node if found, otherwise the default value. Defaults to None
        """
        itr = self.iter()
        match key:
            case Node():
                return first_true(itr, default, lambda node: node == key)
            case str():
                return first_true(itr, default, lambda node: node.name == key)
            case _:
                return first_true(itr, default, key)  # type: ignore

    @overload
    def search_space(
        self,
        parser: Literal["configspace"],
        *,
        flat: bool = False,
        conditionals: bool = True,
        delim: str = ":",
    ) -> ConfigurationSpace:
        ...

    @overload
    def search_space(
        self,
        parser: Literal["optuna"],
        *,
        seed: int | None = None,
        flat: bool = False,
        conditionals: bool = True,
        delim: str = ":",
    ) -> OptunaSearchSpace:
        ...

    @overload
    def search_space(
        self,
        parser: Callable[Concatenate[Node, P], ParserOutput],
        *parser_args: P.args,
        **parser_kwargs: P.kwargs,
    ) -> ParserOutput:
        ...

    def search_space(
        self,
        parser: (
            Callable[Concatenate[Node, P], ParserOutput]
            | Literal["configspace", "optuna"]
        ),
        *parser_args: P.args,
        **parser_kwargs: P.kwargs,
    ) -> ParserOutput | ConfigurationSpace | OptunaSearchSpace:
        """Get the search space for this node.

        Args:
            parser: The parser to use. This can be a function that takes in
                the node and returns the search space or a string that is one of:

                * `#!python "configspace"`: Build a
                    [`ConfigSpace.ConfigurationSpace`](https://automl.github.io/ConfigSpace/master/)
                    out of this node.
                * `#!python "optuna"`: Build a dict of hyperparameters that Optuna can
                    use in its [ask and tell methods](https://optuna.readthedocs.io/en/stable/tutorial/20_recipes/009_ask_and_tell.html#define-and-run)

            parser_args: The positional arguments to pass to the parser
            parser_kwargs: The keyword arguments to pass to the parser

        Returns:
            The search space
        """
        match parser:
            case "configspace":
                from amltk.pipeline.parsers.configspace import parser as cs_parser

                return cs_parser(self, *parser_args, **parser_kwargs)  # type: ignore
            case "optuna":
                from amltk.pipeline.parsers.optuna import parser as optuna_parser

                return optuna_parser(self, *parser_args, **parser_kwargs)  # type: ignore
            case str():  # type: ignore
                raise ValueError(
                    f"Invalid str for parser {parser}. "
                    "Please use 'configspace' or 'optuna' or pass in your own"
                    " parser function",
                )
            case _:
                return parser(self, *parser_args, **parser_kwargs)

    @overload
    def build(
        self,
        builder: Literal["sklearn"],
        *builder_args: Any,
        pipeline_type: type[SklearnPipelineT] = SklearnPipeline,
        **builder_kwargs: Any,
    ) -> SklearnPipelineT:
        ...

    @overload
    def build(
        self,
        builder: Literal["sklearn"],
        *builder_args: Any,
        **builder_kwargs: Any,
    ) -> SklearnPipeline:
        ...

    @overload
    def build(
        self,
        builder: Callable[Concatenate[Node, P], BuilderOutput],
        *builder_args: P.args,
        **builder_kwargs: P.kwargs,
    ) -> BuilderOutput:
        ...

    def build(
        self,
        builder: Callable[Concatenate[Node, P], BuilderOutput] | Literal["sklearn"],
        *builder_args: P.args,
        **builder_kwargs: P.kwargs,
    ) -> BuilderOutput | SklearnPipeline:
        """Build a concrete object out of this node.

        Args:
            builder: The builder to use. This can be a function that takes in
                the node and returns the object or a string that is one of:

                * `#!python "sklearn"`: Build a
                    [`sklearn.pipeline.Pipeline`][sklearn.pipeline.Pipeline]
                    out of this node.

            builder_args: The positional arguments to pass to the builder
            builder_kwargs: The keyword arguments to pass to the builder

        Returns:
            The built object
        """
        match builder:
            case "sklearn":
                from amltk.pipeline.builders.sklearn import build as _build

                return _build(self, *builder_args, **builder_kwargs)  # type: ignore
            case _:
                return builder(self, *builder_args, **builder_kwargs)

    def factorize(
        self,
        *,
        min_depth: int = 0,
        max_depth: int | None = None,
        current_depth: int = 0,
        factor_by: Callable[[Node], bool] | None = None,
        assign_child: Callable[[Node, Node], Node] | None = None,
    ) -> Iterator[Self]:
        """Please see [`factorize()`][amltk.pipeline.ops.factorize]."""  # noqa: D402
        from amltk.pipeline.ops import factorize

        yield from factorize(
            self,
            min_depth=min_depth,
            max_depth=max_depth,
            current_depth=current_depth,
            factor_by=factor_by,
            assign_child=assign_child,
        )

    def _rich_iter(self) -> Iterator[RenderableType]:
        """Iterate the panels for rich printing."""
        yield self.__rich__()
        for node in self.nodes:
            yield from node._rich_iter()

    def _rich_table_items(self) -> Iterator[tuple[RenderableType, ...]]:
        """Get the items to add to the rich table."""
        from rich.pretty import Pretty

        from amltk._richutil import Function

        if self.item is not None:
            if isinstance(self.item, type) or callable(self.item):
                yield "item", Function(self.item, signature="...")
            else:
                yield "item", Pretty(self.item)

        if self.config is not None:
            yield "config", Pretty(self.config)

        if self.space is not None:
            yield "space", Pretty(self.space)

        if self.fidelities is not None:
            yield "fidelity", Pretty(self.fidelities)

        if self.config_transform is not None:
            yield "transform", Function(self.config_transform, signature="...")

        if self.meta is not None:
            yield "meta", Pretty(self.meta)

    def _rich_panel_contents(self) -> Iterator[RenderableType]:
        from rich.table import Table
        from rich.text import Text

        options = self.RICH_OPTIONS

        if panel_contents := list(self._rich_table_items()):
            table = Table.grid(padding=(0, 1), expand=False)
            for tup in panel_contents:
                table.add_row(*tup, style="default")
            table.add_section()
            yield table

        if len(self.nodes) > 0:
            match options:
                case RichOptions(node_orientation="horizontal"):
                    table = Table.grid(padding=(0, 1), expand=False)
                    nodes = [node.__rich__() for node in self.nodes]
                    table.add_row(*nodes)
                    yield table
                case RichOptions(node_orientation="vertical"):
                    first, *rest = self.nodes
                    yield first.__rich__()
                    for node in rest:
                        yield Text("↓", style="bold", justify="center")
                        yield node.__rich__()
                case _:
                    raise ValueError(f"Invalid orientation {options.node_orientation}")

    def display(self, *, full: bool = False) -> RenderableType:
        """Display this node.

        Args:
            full: Whether to display the full node or just a summary
        """
        if not full:
            return self.__rich__()

        from rich.console import Group as RichGroup

        return RichGroup(*self._rich_iter())

    @override
    def __rich__(self) -> Panel:
        from rich.console import Group as RichGroup
        from rich.panel import Panel
        from rich.text import Text

        clr = self.RICH_OPTIONS.panel_color
        title = Text.assemble(
            (classname(self), f"{clr} bold"),
            "(",
            (self.name, f"{clr} italic"),
            ")",
            style="default",
            end="",
        )
        contents = list(self._rich_panel_contents())
        _content = contents[0] if len(contents) == 1 else RichGroup(*contents)
        return Panel(
            _content,
            title=title,
            title_align="left",
            border_style=clr,
            expand=False,
        )

    def _fufill_param_requests(
        self,
        config: Config,
        params: Mapping[str, Any] | None = None,
    ) -> Config:
        _params = params or {}
        new_config = dict(config)

        for k, request in config.items():
            match request:
                case ParamRequest(key=request_key) if request_key in _params:
                    new_config[k] = _params[request_key]
                case ParamRequest(default=default) if request.has_default:
                    new_config[k] = default
                case ParamRequest():
                    raise RequestNotMetError(
                        f"Missing {request=} for {self}.\nparams={params}",
                    )
                case _:
                    continue

        return new_config

    def register_optimization_loop(  # noqa: C901, PLR0915, PLR0912
        self,
        target: (
            EvaluationProtocol
            | Callable[[Trial, Node], Trial.Report]
            | Task[[Trial, Node], Trial.Report]
        ),
        metric: Metric | Sequence[Metric],
        *,
        optimizer: type[Optimizer] | Optimizer.CreateSignature | None = None,
        seed: Seed | None = None,
        max_trials: int | None = 3,
        n_workers: int = 1,
        working_dir: str | Path | PathBucket | None = None,
        scheduler: Scheduler | None = None,
        history: History | None = None,
        on_begin: OnBeginCallbackSignature | None = None,
        on_trial_exception: Literal["raise", "end", "continue"] = "raise",
        # Plugin creating arguments
        plugins: Plugin | Iterable[Plugin] | None = None,
        process_memory_limit: int | tuple[int, str] | None = None,
        process_walltime_limit: int | tuple[float, str] | None = None,
        process_cputime_limit: int | tuple[float, str] | None = None,
        threadpool_limit_ctl: bool | int | None = None,
    ) -> Scheduler:
        """Setup a pipeline to be optimized in a loop.

        Args:
            target:
                The function against which to optimize.

                * If `target` is an
                [`EvaluationProtocol`][amltk.evalutors.evaluation_protocol.EvaluationProtocol],
                then it will be used to evaluate the pipeline.

                * If `target` is a function, then it must take in a
                [`Trial`][amltk.optimization.trial.Trial] as the first argument
                and a [`Node`][amltk.pipeline.node.Node] second argument, returning a
                [`Trial.Report`][amltk.optimization.trial.Trial.Report]. Please refer to
                the [optimization guide](../../../guides/optimization.md) for more.

                * If `target` is a [`Task`][amltk.scheduling.task.Task], then
                this is not implemeneted yet. Sorry
            metric:
                The metric(s) that will be passed to `optimizer=`. These metrics
                should align with what is being computed in `target=`.
            optimizer:
                The optimizer to use. If `None`, then AMLTK will go through a list
                of known optimizers and use the first one it can find which was installed.

                Alternatively, this can be a class inheriting from
                [`Optimizer`][amltk.optimization.optimizer.Optimizer] or else
                a signature match [`Optimizer.CreateSignature`][amltk.optimization.Optimizer.CreateSignature]

                ??? tip "`Optimizer.CreateSignature`"

                    ::: amltk.optimization.Optimizer.CreateSignature

            seed:
                A [`seed`][amltk.types.Seed] for the optimizer to use.
            n_workers:
                The numer of workers to use to evaluate this pipeline.
                If no `scheduler=` is provided, then one will be created for
                you as [`Scheduler.with_processes(n_workers)`][amltk.scheduling.Scheduler.with_processes].
                If you provide your own `scheduler=` then this will limit the maximum
                amount of concurrent trials for this pipeline that will be evaluating
                at once.
            working_dir:
                A working directory to use for the optimizer and the trials.
                Any items you store in trials will be located in this directory,
                where the [`trial.name`][amltk.optimization.Trial.name] will be
                used as a subfolder where any contents stored with
                [`trial.store()`][amltk.optimization.trial.Trial.store] will be put there.
                Please see the [optimization guide](../../../guides/optimization.md)
                for more on trial storage.
            scheduler:
                The specific [`Scheduler`][amltk.scheduling.Scheduler] to use.
                If `None`, then one will be created for you with
                [`Scheduler.with_processes(n_workers)`][amltk.scheduling.Scheduler.with_processes]
            history:
                A [`History`][amltk.optimization.history.History] to store the
                [`Trial.Report`][amltk.optimization.Trial.Report]s in. You
                may pass in your own if you wish for this method to store
                it there instead of creating its own.
            on_begin:
                A callback that will be called before the scheduler is run. This
                can be used to hook into the life-cycle of the optimization and
                perform custom routines. Please see the
                [scheduling guide](../../../guides/scheduling.md) for more.

                ??? tip "on_begin signature"

                    ::: amltk.pipeline.node.OnBeginCallbackSignature

            on_trial_exception:
                What to do when a trial returns a fail report from
                [`trial.fail()`][amltk.optimization.trial.Trial.fail] or
                [`trial.crashed()`][amltk.optimization.trial.Trial.crashed]
                that contains an exception.

                Please see the [optimization guide](../../../guides/optimization.md)
                for more. In all cases, the exception will be attached to the
                [`Trial.Report`][amltk.optimization.Trial.Report] object under
                [`report.exception`][amltk.optimization.Trial.Report.exception].

                * If `#!python "raise"`, then the exception will be raised
                immediatly and the optimization process will halt. The default
                and good for initial development.
                * If `#!python "end"`, then the exception will be caught and
                the optimization process will end gracefully.
                * If `#!python "continue"`, the exception will be ignored and
                the optimization procedure will continue.

            max_trials:
                The maximum number of trials to run. If `None`, then the
                optimization will continue for as long as the scheduler is
                running. You'll likely want to configure this.
            process_memory_limit:
                If specified, the [`Task`][amltk.scheduling.task.Task] will
                use the
                [`PynisherPlugin`][amltk.scheduling.plugins.pynisher.PynisherPlugin]
                to limit the memory the process can use. Please
                refer to the
                [plugins `pynisher` reference](../../../reference/scheduling/plugins.md#pynisher)
                for more as there are platform limitations and additional
                dependancies required.
            process_walltime_limit:
                If specified, the [`Task`][amltk.scheduling.task.Task] will
                use the
                [`PynisherPlugin`][amltk.scheduling.plugins.pynisher.PynisherPlugin]
                to limit the wall time the process can use. Please
                refer to the
                [plugins `pynisher` reference](../../../reference/scheduling/plugins.md#pynisher)
                for more as there are platform limitations and additional
                dependancies required.
            process_cputime_limit:
                If specified, the [`Task`][amltk.scheduling.task.Task] will
                use the
                [`PynisherPlugin`][amltk.scheduling.plugins.pynisher.PynisherPlugin]
                to limit the cputime the process can use. Please
                refer to the
                [plugins `pynisher` reference](../../../reference/scheduling/plugins.md#pynisher)
                for more as there are platform limitations and additional
                dependancies required.
            threadpool_limit_ctl:
                If specified, the [`Task`][amltk.scheduling.task.Task] will
                use the
                [`ThreadPoolCTLPlugin`][amltk.scheduling.plugins.threadpoolctl.ThreadPoolCTLPlugin]
                to limit the number of threads used by compliant libraries.
                **Notably**, this includes scikit-learn, for which running multiple
                in parallel can be problematic if not adjusted accordingly.

                The default behavior (when `None`) is to auto-detect whether this
                is applicable. This is done by checking if `sklearn` is installed
                and if the first node in the pipeline has a `BaseEstimator` item.
                Please set this to `True`/`False` depending on your preference.
            plugins:
                Additional plugins to attach to the eventual
                [`Task`][amltk.scheduling.task.Task] that will be executed by
                the [`Scheduler`][amltk.scheduling.Scheduler]. Please
                refer to the
                [plugins reference](../../../reference/scheduling/plugins.md) for more.
        """  # noqa: E501
        match history:
            case None:
                history = History()
            case History():
                pass
            case _:
                raise ValueError(f"Invalid history {history}. Must be a History")

        _plugins: tuple[Plugin, ...]
        match plugins:
            case None:
                _plugins = ()
            case Plugin():
                _plugins = (plugins,)
            case Iterable():
                _plugins = tuple(plugins)
            case _:
                raise ValueError(
                    f"Invalid plugins {plugins}. Must be a Plugin or an Iterable of"
                    " Plugins",
                )

        if any(
            limit is not None
            for limit in (
                process_memory_limit,
                process_walltime_limit,
                process_cputime_limit,
            )
        ):
            try:
                from amltk.scheduling.plugins.pynisher import PynisherPlugin
            except ImportError as e:
                raise ImportError(
                    "You must install `pynisher` to use `trial_*_limit`"
                    " You can do so with `pip install amltk[pynisher]`"
                    " or `pip install pynisher` directly",
                ) from e
            # TODO: I'm hesitant to add even more arguments to the `optimize`
            # signature, specifically for `mp_context`.
            plugin = PynisherPlugin(
                memory_limit=process_memory_limit,
                walltime_limit=process_walltime_limit,
                cputime_limit=process_cputime_limit,
            )
            _plugins = (*_plugins, plugin)

        # If threadpool_limit_ctl None, we should default to inspecting if it's
        # an sklearn pipeline. This is because sklearn pipelines
        # run in parallel will over-subscribe the CPU and cause
        # the system to slow down.
        # We use a heuristic to check this by checking if the item at the head
        # of this node is a subclass of sklearn.base.BaseEstimator
        match threadpool_limit_ctl:
            case None:
                from amltk._util import threadpoolctl_heuristic

                threadpool_limit_ctl = False
                if threadpoolctl_heuristic(self.item):
                    threadpool_limit_ctl = 1
                    warnings.warn(
                        "Detected an sklearn pipeline. Setting `threadpool_limit_ctl`"
                        " to True. This will limit the number of threads spawned by"
                        " sklearn to the number of cores on the machine. This is"
                        " because sklearn pipelines run in parallel will over-subscribe"
                        " the CPU and cause the system to slow down."
                        "\nPlease set `threadpool_limit_ctl=False` if you do not want"
                        " this behaviour and set it to `True` to silence this warning.",
                        stacklevel=2,
                    )
            case True:
                threadpool_limit_ctl = 1
            case False:
                pass
            case int():
                pass
            case _:
                raise ValueError(
                    f"Invalid threadpool_limit_ctl {threadpool_limit_ctl}."
                    " Must be a bool or an int",
                )

        if threadpool_limit_ctl is not False:
            from amltk.scheduling.plugins.threadpoolctl import ThreadPoolCTLPlugin

            _plugins = (*_plugins, ThreadPoolCTLPlugin(threadpool_limit_ctl))

        match max_trials:
            case None:
                pass
            case int() if max_trials > 0:
                from amltk.scheduling.plugins import Limiter

                _plugins = (*_plugins, Limiter(max_calls=max_trials))
            case _:
                raise ValueError(f"{max_trials=} must be a positive int")

        from amltk.evalutors.evaluation_protocol import EvaluationProtocol

        match target:
            case EvaluationProtocol():
                pass
            case Task():  # type: ignore # NOTE not sure why pyright complains here
                # TODO: When updating this, please update the docstring too
                raise NotImplementedError(
                    "Not sure how to handle an already created task yet",
                )
            case _ if callable(target):
                from amltk.evalutors.evaluation_protocol import CustomProtocol

                target = CustomProtocol(target)
            case _:
                raise ValueError(
                    f"Invalid target {target}. Must be a function or an"
                    " EvaluationProtocol",
                )

        from amltk.scheduling.scheduler import Scheduler

        match scheduler:
            case None:
                scheduler = Scheduler.with_processes(n_workers)
            case Scheduler():
                pass
            case _:
                raise ValueError(f"Invalid scheduler {scheduler}. Must be a Scheduler")

        # NOTE: I'm not particularly fond of this hack but I assume most people
        # when prototyping don't care for the actual underlying optimizer and
        # so we should just *pick one*.
        create_optimizer: Optimizer.CreateSignature
        match optimizer:
            case None:
                first_opt_class = next(
                    Optimizer._get_known_importable_optimizer_classes(),
                    None,
                )
                if first_opt_class is None:
                    raise ValueError(
                        "No optimizer was given and no known importable optimizers were"
                        " found. Please consider giving one explicitly or  installing"
                        " one of the following packages:\n"
                        "\n - optuna"
                        "\n - smac"
                        "\n - neural-pipeline-search",
                    )

                create_optimizer = first_opt_class.create
                opt_name = classname(first_opt_class)
            case type():
                if not issubclass(optimizer, Optimizer):
                    raise ValueError(
                        f"Invalid optimizer {optimizer}. Must be a subclass of"
                        " Optimizer or a function that returns an Optimizer",
                    )
                create_optimizer = optimizer.create
                opt_name = classname(optimizer)
            case _:
                assert not isinstance(optimizer, type)
                create_optimizer = optimizer
                opt_name = funcname(optimizer)

        match working_dir:
            case None:
                now = datetime.utcnow().isoformat()

                working_dir = PathBucket(f"{opt_name}-{self.name}-{now}")
            case str() | Path():
                working_dir = PathBucket(working_dir)
            case PathBucket():
                pass
            case _:
                raise ValueError(
                    f"Invalid working_dir {working_dir}."
                    " Must be a str, Path or PathBucket",
                )

        _optimizer = create_optimizer(
            space=self,
            metrics=metric,
            bucket=working_dir,
            seed=seed,
        )

        task = target.task(scheduler=scheduler, plugins=_plugins)

        if on_begin is not None:
            hook = partial(on_begin, task, scheduler, history)
            scheduler.on_start(hook)

        @scheduler.on_start
        def launch_initial_trials() -> None:
            trials = _optimizer.ask(n=n_workers)
            for trial in trials:
                task.submit(trial, self)

        from amltk.optimization.trial import Trial

        @task.on_result
        def add_report_to_history(_: Any, report: Trial.Report) -> None:
            history.add(report)
            match report.status:
                case Trial.Status.SUCCESS:
                    return
                case Trial.Status.FAIL | Trial.Status.CRASHED | Trial.Status.UNKNOWN:
                    match on_trial_exception:
                        case "raise":
                            if report.exception is None:
                                raise RuntimeError(
                                    f"Trial finished with status {report.status} but"
                                    " no exception was attached!",
                                )
                            raise report.exception
                        case "end":
                            scheduler.stop(
                                stop_msg=f"Trial finished with status {report.status}",
                                exception=report.exception,
                            )
                        case "continue":
                            pass
                case _:
                    raise ValueError(f"Invalid status {report.status}")

        @task.on_result
        def run_next_trial(*_: Any) -> None:
            if scheduler.running():
                trial = _optimizer.ask()
                task.submit(trial, self)

        return scheduler

    def optimize(
        self,
        target: (
            EvaluationProtocol
            | Callable[[Trial, Node], Trial.Report]
            | Task[[Trial, Node], Trial.Report]
        ),
        metric: Metric | Sequence[Metric],
        *,
        optimizer: type[Optimizer] | Optimizer.CreateSignature | None = None,
        seed: Seed | None = None,
        max_trials: int | None = 3,
        n_workers: int = 1,
        timeout: float | None = None,
        working_dir: str | Path | PathBucket | None = None,
        scheduler: Scheduler | None = None,
        history: History | None = None,
        on_begin: OnBeginCallbackSignature | None = None,
        on_trial_exception: Literal["raise", "end", "continue"] = "raise",
        # Plugin creating arguments
        plugins: Plugin | Iterable[Plugin] | None = None,
        process_memory_limit: int | tuple[int, str] | None = None,
        process_walltime_limit: int | tuple[float, str] | None = None,
        process_cputime_limit: int | tuple[float, str] | None = None,
        threadpool_limit_ctl: bool | int | None = None,
        # `scheduler.run()` arguments
        display: bool | Literal["auto"] = "auto",
        wait: bool = True,
        on_scheduler_exception: Literal["raise", "end", "continue"] = "raise",
    ) -> History:
        """Optimize a pipeline on a given target function or evaluation protocol.

        Args:
            target:
                The function against which to optimize.

                * If `target` is an
                [`EvaluationProtocol`][amltk.evalutors.evaluation_protocol.EvaluationProtocol],
                then it will be used to evaluate the pipeline.

                * If `target` is a function, then it must take in a
                [`Trial`][amltk.optimization.trial.Trial] as the first argument
                and a [`Node`][amltk.pipeline.node.Node] second argument, returning a
                [`Trial.Report`][amltk.optimization.trial.Trial.Report]. Please refer to
                the [optimization guide](../../../guides/optimization.md) for more.

                * If `target` is a [`Task`][amltk.scheduling.task.Task], then
                this is not implemeneted yet. Sorry
            metric:
                The metric(s) that will be passed to `optimizer=`. These metrics
                should align with what is being computed in `target=`.
            optimizer:
                The optimizer to use. If `None`, then AMLTK will go through a list
                of known optimizers and use the first one it can find which was installed.

                Alternatively, this can be a class inheriting from
                [`Optimizer`][amltk.optimization.optimizer.Optimizer] or else
                a signature match [`Optimizer.CreateSignature`][amltk.optimization.Optimizer.CreateSignature]

                ??? tip "`Optimizer.CreateSignature`"

                    ::: amltk.optimization.Optimizer.CreateSignature

            seed:
                A [`seed`][amltk.types.Seed] for the optimizer to use.
            n_workers:
                The numer of workers to use to evaluate this pipeline.
                If no `scheduler=` is provided, then one will be created for
                you as [`Scheduler.with_processes(n_workers)`][amltk.scheduling.Scheduler.with_processes].
                If you provide your own `scheduler=` then this will limit the maximum
                amount of concurrent trials for this pipeline that will be evaluating
                at once.
            timeout:
                How long to run the scheduler for. This parameter only takes
                effect if `setup_only=False` which is the default. Otherwise,
                it will be ignored.
            display:
                Whether to display the scheduler during running. By default
                it is `"auto"` which means to enable the display if running
                in a juptyer notebook or colab. Otherwise, it will be
                `False`.

                This may work poorly if including print statements or logging.
            wait:
                Whether to wait for the scheduler to finish all pending jobs
                if was stopped for any reason, e.g. a `timeout=` or
                [`scheduler.stop()`][amltk.scheduling.Scheduler.stop] was called.
            on_scheduler_exception:
                What to do if an exception occured, either in the submitted task,
                the callback, or any other unknown source during the loop.

                * If `#!python "raise"`, then the exception will be raised
                immediatly and the optimization process will halt. The default
                behavior and good for initial development.
                * If `#!python "end"`, then the exception will be caught and
                the optimization process will end gracefully.
                * If `#!python "continue"`, the exception will be ignored and
                the optimization procedure will continue.
            working_dir:
                A working directory to use for the optimizer and the trials.
                Any items you store in trials will be located in this directory,
                where the [`trial.name`][amltk.optimization.Trial.name] will be
                used as a subfolder where any contents stored with
                [`trial.store()`][amltk.optimization.trial.Trial.store] will be put there.
                Please see the [optimization guide](../../../guides/optimization.md)
                for more on trial storage.
            scheduler:
                The specific [`Scheduler`][amltk.scheduling.Scheduler] to use.
                If `None`, then one will be created for you with
                [`Scheduler.with_processes(n_workers)`][amltk.scheduling.Scheduler.with_processes]
            history:
                A [`History`][amltk.optimization.history.History] to store the
                [`Trial.Report`][amltk.optimization.Trial.Report]s in. You
                may pass in your own if you wish for this method to store
                it there instead of creating its own.
            on_begin:
                A callback that will be called before the scheduler is run. This
                can be used to hook into the life-cycle of the optimization and
                perform custom routines. Please see the
                [scheduling guide](../../../guides/scheduling.md) for more.

                ??? tip "on_begin signature"

                    ::: amltk.pipeline.node.OnBeginCallbackSignature

            on_trial_exception:
                What to do when a trial returns a fail report from
                [`trial.fail()`][amltk.optimization.trial.Trial.fail] or
                [`trial.crashed()`][amltk.optimization.trial.Trial.crashed]
                that contains an exception.

                Please see the [optimization guide](../../../guides/optimization.md)
                for more. In all cases, the exception will be attached to the
                [`Trial.Report`][amltk.optimization.Trial.Report] object under
                [`report.exception`][amltk.optimization.Trial.Report.exception].

                * If `#!python "raise"`, then the exception will be raised
                immediatly and the optimization process will halt. The default
                and good for initial development.
                * If `#!python "end"`, then the exception will be caught and
                the optimization process will end gracefully.
                * If `#!python "continue"`, the exception will be ignored and
                the optimization procedure will continue.

            max_trials:
                The maximum number of trials to run. If `None`, then the
                optimization will continue for as long as the scheduler is
                running. You'll likely want to configure this.

            process_memory_limit:
                If specified, the [`Task`][amltk.scheduling.task.Task] will
                use the
                [`PynisherPlugin`][amltk.scheduling.plugins.pynisher.PynisherPlugin]
                to limit the memory the process can use. Please
                refer to the
                [plugins `pynisher` reference](../../../reference/scheduling/plugins.md#pynisher)
                for more as there are platform limitations and additional
                dependancies required.
            process_walltime_limit:
                If specified, the [`Task`][amltk.scheduling.task.Task] will
                use the
                [`PynisherPlugin`][amltk.scheduling.plugins.pynisher.PynisherPlugin]
                to limit the wall time the process can use. Please
                refer to the
                [plugins `pynisher` reference](../../../reference/scheduling/plugins.md#pynisher)
                for more as there are platform limitations and additional
                dependancies required.
            process_cputime_limit:
                If specified, the [`Task`][amltk.scheduling.task.Task] will
                use the
                [`PynisherPlugin`][amltk.scheduling.plugins.pynisher.PynisherPlugin]
                to limit the cputime the process can use. Please
                refer to the
                [plugins `pynisher` reference](../../../reference/scheduling/plugins.md#pynisher)
                for more as there are platform limitations and additional
                dependancies required.
            threadpool_limit_ctl:
                If specified, the [`Task`][amltk.scheduling.task.Task] will
                use the
                [`ThreadPoolCTLPlugin`][amltk.scheduling.plugins.threadpoolctl.ThreadPoolCTLPlugin]
                to limit the number of threads used by compliant libraries.
                **Notably**, this includes scikit-learn, for which running multiple
                in parallel can be problematic if not adjusted accordingly.

                The default behavior (when `None`) is to auto-detect whether this
                is applicable. This is done by checking if `sklearn` is installed
                and if the first node in the pipeline has a `BaseEstimator` item.
                Please set this to `True`/`False` depending on your preference.
            plugins:
                Additional plugins to attach to the eventual
                [`Task`][amltk.scheduling.task.Task] that will be executed by
                the [`Scheduler`][amltk.scheduling.Scheduler]. Please
                refer to the
                [plugins reference](../../../reference/scheduling/plugins.md) for more.
        """  # noqa: E501
        match history:
            case None:
                history = History()
            case History():
                pass
            case _:
                raise ValueError(f"Invalid history {history}. Must be a History")

        scheduler = self.register_optimization_loop(
            target=target,
            metric=metric,
            optimizer=optimizer,
            seed=seed,
            max_trials=max_trials,
            n_workers=n_workers,
            working_dir=working_dir,
            scheduler=scheduler,
            history=history,
            on_begin=on_begin,
            on_trial_exception=on_trial_exception,
            plugins=plugins,
            process_memory_limit=process_memory_limit,
            process_walltime_limit=process_walltime_limit,
            process_cputime_limit=process_cputime_limit,
            threadpool_limit_ctl=threadpool_limit_ctl,
        )
        scheduler.run(
            wait=wait,
            timeout=timeout,
            on_exception=on_scheduler_exception,
            display=display,
        )
        return history



"""A [`Metric`][amltk.optimization.Metric] to let optimizers know how to
handle numeric values properly.

A `Metric` is defined by a `.name: str` and whether it is better to `.minimize: bool`
the metric. Further, you can specify `.bounds: tuple[lower, upper]` which can
help optimizers and other code know how to treat metrics.

To easily convert between [`loss`][amltk.optimization.Metric.Value.loss],
[`score`][amltk.optimization.Metric.Value.score] of a
a value in a [`Metric.Value`][amltk.optimization.Metric.Value] object.

If the metric is bounded, you can also get the
[`distance_to_optimal`][amltk.optimization.Metric.Value.distance_to_optimal]
which is the distance to the optimal value.

```python exec="true" source="material-block" result="python"
from amltk.optimization import Metric

acc = Metric("accuracy", minimize=False, bounds=(0.0, 1.0))

acc_value = acc.as_value(0.9)
print(f"Cost: {acc_value.distance_to_optimal}")  # Distance to optimal.
print(f"Loss: {acc_value.loss}")  # Something that can be minimized
print(f"Score: {acc_value.score}")  # Something that can be maximized
```

"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing_extensions import Self, override


@dataclass(frozen=True)
class Metric:
    """A metric with a given name, optimal direction, and possible bounds."""

    name: str
    """The name of the metric."""

    minimize: bool = field(kw_only=True, default=True)
    """Whether to minimize or maximize the metric."""

    bounds: tuple[float, float] | None = field(kw_only=True, default=None)
    """The bounds of the metric, if any."""

    def __post_init__(self) -> None:
        if self.bounds is not None:
            lower, upper = self.bounds
            if lower > upper:
                raise ValueError(f"Lower bound {lower} > upper bound {upper}")

            object.__setattr__(self, "bounds", (float(lower), float(upper)))

        if self.name[0].isdigit():
            raise ValueError(
                f"Metric name {self.name} cannot start with a digit."
                " Must be a valid Python identifier.",
            )

        for c in "[](){}<>|&^%$#@!~`":
            if c in self.name:
                raise ValueError(
                    f"Metric name {self.name} cannot contain '{c}'."
                    " Must be a valid Python identifier.",
                )

    @override
    def __str__(self) -> str:
        parts = [self.name]
        if self.bounds is not None:
            parts.append(f"[{self.bounds[0]}, {self.bounds[1]}]")
        parts.append(f"({'minimize' if self.minimize else 'maximize'})")
        return " ".join(parts)

    @classmethod
    def from_str(cls, s: str) -> Self:
        """Create an metric from a str.

        ```python exec="true" source="material-block" result="python"
        from amltk.optimization import Metric

        s = "loss (minimize)"
        metric = Metric.from_str(s)
        print(metric)

        s = "accuracy [0.0, 1.0] (maximize)"
        metric = Metric.from_str(s)
        print(metric)
        ```

        Args:
            s: The string to parse.

        Returns:
            The parsed metric.
        """
        splits = s.split(" ")
        # No bounds
        if len(splits) == 2:  # noqa: PLR2004
            name, minimize_str = splits
            bounds = None
        else:
            name, lower_str, upper_str, minimize_str = splits
            bounds = (float(lower_str[1:-1]), float(upper_str[:-1]))

        minimize = minimize_str == "(minimize)"
        return cls(name=name, minimize=minimize, bounds=bounds)

    @property
    def worst(self) -> Metric.Value:
        """The worst possible value of the metric."""
        if self.bounds:
            v = self.bounds[1] if self.minimize else self.bounds[0]
            return self.as_value(v)

        v = float("inf") if self.minimize else float("-inf")
        return self.as_value(v)

    @property
    def optimal(self) -> Metric.Value:
        """The optimal value of the metric."""
        if self.bounds:
            v = self.bounds[0] if self.minimize else self.bounds[1]
            return self.as_value(v)
        v = float("-inf") if self.minimize else float("inf")
        return self.as_value(v)

    def as_value(self, value: float | int) -> Metric.Value:
        """Convert a value to an metric value."""
        return Metric.Value(metric=self, value=float(value))

    def __call__(self, value: float | int) -> Metric.Value:
        """Convert a value to an metric value."""
        return Metric.Value(metric=self, value=float(value))

    @dataclass(frozen=True, order=True)
    class Value:
        """A recorded value of an metric."""

        metric: Metric = field(compare=False, hash=True)
        """The metric."""

        value: float = field(compare=True, hash=True)
        """The value of the metric."""

        @property
        def minimize(self) -> bool:
            """Whether to minimize or maximize the metric."""
            return self.metric.minimize

        @property
        def bounds(self) -> tuple[float, float] | None:
            """Whether to minimize or maximize the metric."""
            return self.metric.bounds

        @property
        def name(self) -> str:
            """The name of the metric."""
            return self.metric.name

        @property
        def loss(self) -> float:
            """Convert a value to a loss."""
            if self.minimize:
                return float(self.value)
            return -float(self.value)

        @property
        def score(self) -> float:
            """Convert a value to a score."""
            if self.minimize:
                return -float(self.value)
            return float(self.value)

        @property
        def distance_to_optimal(self) -> float | None:
            """The distance to the optimal value, using the bounds if possible."""
            match self.bounds:
                case None:
                    return None
                case (lower, upper) if lower <= self.value <= upper:
                    if self.minimize:
                        return abs(self.value - lower)
                    return abs(self.value - upper)
                case (lower, upper):
                    raise ValueError(f"Value {self.value} is not within {self.bounds=}")

            return None

        def __float__(self) -> float:
            """Convert a value to a float."""
            return float(self.value)

        @override
        def __eq__(self, __value: object) -> bool:
            """Check if two values are equal."""
            if isinstance(__value, Metric.Value):
                return self.value == __value.value
            if isinstance(__value, float | int):
                return self.value == float(__value)
            return NotImplemented



from amltk.scheduling.events import Emitter, Event, Subscriber
from amltk.scheduling.executors import SequentialExecutor
from amltk.scheduling.plugins import Comm, Limiter, Plugin
from amltk.scheduling.scheduler import ExitState, Scheduler
from amltk.scheduling.task import Task

__all__ = [
    "Scheduler",
    "Comm",
    "Task",
    "SequentialExecutor",
    "Plugin",
    "Limiter",
    "ExitState",
    "Comm",
    "Emitter",
    "Subscriber",
    "Event",
]



"""Module containing the base protocol of a loader.

For concrete implementations based on the `key` being
a [`Path`][pathlib.Path] see the
[`path_loaders`][amltk.store.paths.path_loaders.PathLoader]
module.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Generic, TypeVar

T = TypeVar("T")
KeyT_contra = TypeVar("KeyT_contra", contravariant=True)


class Loader(ABC, Generic[KeyT_contra, T]):
    """The base definition of a Loader.

    A Loader is a class that can save and load objects to and from a
    bucket. The Loader is responsible for knowing how to save and load
    objects of a particular type at a given key.
    """

    name: ClassVar[str]
    """The name of the loader."""

    @classmethod
    @abstractmethod
    def can_load(cls, key: KeyT_contra, /, *, check: type[T] | None = None) -> bool:
        """Return True if this loader supports the resource at key.

        This is used to determine which loader to use when loading a
        resource from a key.

        Args:
            key: The key used to identify the resource
            check: If the loader can support loading a specific type
                of object.
        """
        ...

    @classmethod
    @abstractmethod
    def can_save(cls, obj: Any, key: KeyT_contra, /) -> bool:
        """Return True if this loader can save this object.

        This is used to determine which loader to use when loading a
        resource from a key.

        Args:
            obj: The object to save.
            key: The key used to identify the resource
        """
        ...

    @classmethod
    @abstractmethod
    def save(cls, obj: Any, key: KeyT_contra, /) -> None:
        """Save an object to under the given key.

        Args:
            obj: The object to save.
            key: The key to save the object under.
        """
        ...

    @classmethod
    @abstractmethod
    def load(cls, key: KeyT_contra, /) -> T:
        """Load an object from the given key.

        Args:
            key: The key to load the object from.

        Returns:
            The loaded object.
        """
        ...



"""The [`Limiter`][amltk.scheduling.plugins.Limiter] can limit the number of
times a function is called, how many concurrent instances of it can be running,
or whether it can run while another task is running.

The functionality of the `Limiter` could also be implemented without a plugin but
it gives some nice utility.

??? tip "Usage"

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Limiter

    def fn(x: int) -> int:
        return x + 1

    scheduler = Scheduler.with_processes(1)

    task = scheduler.task(fn, plugins=[Limiter(max_calls=2)])

    @task.on("call-limit-reached")
    def callback(task: Task, *args, **kwargs):
        pass
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```

??? example "`@events`"

    === "`@call-limit-reached`"

        ::: amltk.scheduling.plugins.Limiter.CALL_LIMIT_REACHED

    === "`@concurrent-limit-reached`"

        ::: amltk.scheduling.plugins.Limiter.CONCURRENT_LIMIT_REACHED

    === "`@disabled-due-to-running-task`"

        ::: amltk.scheduling.plugins.Limiter.DISABLED_DUE_TO_RUNNING_TASK
"""
from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar
from typing_extensions import ParamSpec, Self, override

from amltk.scheduling.events import Event
from amltk.scheduling.plugins.plugin import Plugin

if TYPE_CHECKING:
    from rich.panel import Panel

    from amltk.scheduling.task import Task

P = ParamSpec("P")
R = TypeVar("R")
TrialInfo = TypeVar("TrialInfo")


class Limiter(Plugin):
    """A plugin that limits the submission of a task.

    Adds three new events to the task:

    * [`@call-limit-reached`][amltk.scheduling.plugins.Limiter.CALL_LIMIT_REACHED]
    * [`@concurrent-limit-reached`][amltk.scheduling.plugins.Limiter.CONCURRENT_LIMIT_REACHED]
    * [`@disabled-due-to-running-task`][amltk.scheduling.plugins.Limiter.DISABLED_DUE_TO_RUNNING_TASK]
    """  # noqa: E501

    name: ClassVar = "limiter"
    """The name of the plugin."""

    CALL_LIMIT_REACHED: Event[...] = Event("call-limit-reached")
    """The event emitted when the task has reached its call limit.

    Will call any subscribers with the task as the first argument,
    followed by the arguments and keyword arguments that were passed to the task.

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Limiter

    def fn(x: int) -> int:
        return x + 1

    scheduler = Scheduler.with_processes(1)

    task = scheduler.task(fn, plugins=[Limiter(max_calls=2)])

    @task.on("call-limit-reached")
    def callback(task: Task, *args, **kwargs):
        pass
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    CONCURRENT_LIMIT_REACHED: Event[...] = Event("concurrent-limit-reached")
    """The event emitted when the task has reached its concurrent call limit.

    Will call any subscribers with the task as the first argument, followed by the
    arguments and keyword arguments that were passed to the task.

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Limiter

    def fn(x: int) -> int:
        return x + 1

    scheduler = Scheduler.with_processes(2)

    task = scheduler.task(fn, plugins=[Limiter(max_concurrent=2)])

    @task.on("concurrent-limit-reached")
    def callback(task: Task, *args, **kwargs):
        pass
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    DISABLED_DUE_TO_RUNNING_TASK: Event[...] = Event("disabled-due-to-running-task")
    """The event emitter when the task was not submitted due to some other
    running task.

    Will call any subscribers with the task as first argument, followed by
    the arguments and keyword arguments that were passed to the task.

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import Limiter

    def fn(x: int) -> int:
        return x + 1

    scheduler = Scheduler.with_processes(2)

    other_task = scheduler.task(fn)
    task = scheduler.task(fn, plugins=[Limiter(not_while_running=other_task)])

    @task.on("disabled-due-to-running-task")
    def callback(task: Task, *args, **kwargs):
        pass
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    def __init__(
        self,
        *,
        max_calls: int | None = None,
        max_concurrent: int | None = None,
        not_while_running: Task | Iterable[Task] | None = None,
    ):
        """Initialize the plugin.

        Args:
            max_calls: The maximum number of calls to the task.
            max_concurrent: The maximum number of calls of this task that can
                be in the queue.
            not_while_running: A task or iterable of tasks that if active, will prevent
                this task from being submitted.
        """
        super().__init__()

        if not_while_running is None:
            not_while_running = []
        elif isinstance(not_while_running, Iterable):
            not_while_running = list(not_while_running)
        else:
            not_while_running = [not_while_running]

        self.max_calls = max_calls
        self.max_concurrent = max_concurrent
        self.not_while_running = not_while_running
        self.task: Task | None = None

        if isinstance(max_calls, int) and not max_calls > 0:
            raise ValueError("max_calls must be greater than 0")

        if isinstance(max_concurrent, int) and not max_concurrent > 0:
            raise ValueError("max_concurrent must be greater than 0")

        self._calls = 0

    @override
    def attach_task(self, task: Task) -> None:
        """Attach the plugin to a task."""
        self.task = task

        if self.task in self.not_while_running:
            raise ValueError(
                f"Task {self.task} was found in the {self.not_while_running=}"
                " list. This is disabled but please raise an issue if you think this"
                " has sufficient use case.",
            )

        task.emitter.add_event(
            self.CALL_LIMIT_REACHED,
            self.CONCURRENT_LIMIT_REACHED,
            self.DISABLED_DUE_TO_RUNNING_TASK,
        )

        # Make sure to increment the count when a task was submitted
        task.on_submitted(self._increment_call_count, hidden=True)

    @property
    def n_running(self) -> int:
        """Return the number of running tasks."""
        assert self.task is not None
        return len(self.task.queue)

    @override
    def pre_submit(
        self,
        fn: Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> tuple[Callable[P, R], tuple, dict] | None:
        """Pre-submit hook.

        Prevents submission of the task if it exceeds any of the set limits.
        """
        assert self.task is not None

        if self.max_calls is not None and self._calls >= self.max_calls:
            self.task.emitter.emit(self.CALL_LIMIT_REACHED, self.task, *args, **kwargs)
            return None

        if self.max_concurrent is not None and self.n_running >= self.max_concurrent:
            self.task.emitter.emit(
                self.CONCURRENT_LIMIT_REACHED,
                self.task,
                *args,
                **kwargs,
            )
            return None

        for other_task in self.not_while_running:
            if other_task.running():
                self.task.emitter.emit(
                    self.DISABLED_DUE_TO_RUNNING_TASK,
                    other_task,
                    self.task,
                    *args,
                    **kwargs,
                )
                return None

        return fn, args, kwargs

    @override
    def copy(self) -> Self:
        """Return a copy of the plugin."""
        return self.__class__(
            max_calls=self.max_calls,
            max_concurrent=self.max_concurrent,
        )

    def _increment_call_count(self, *_: Any, **__: Any) -> None:
        self._calls += 1

    @override
    def __rich__(self) -> Panel:
        from rich.panel import Panel
        from rich.table import Table
        from rich.text import Text

        from amltk._richutil import Function

        table = Table.grid(padding=(0, 1))
        if self.max_calls is not None:
            table.add_row("Calls", f"{self._calls}/{self.max_calls}")

        if self.max_concurrent is not None:
            table.add_row("Concurrent", f"{self.n_running}/{self.max_concurrent}")

        for task in self.not_while_running:
            f = Function(
                task.function,
                signature="...",
                link=False,
            )
            if task.running():
                table.add_row(
                    "Not While",
                    f,
                    Text(task.unique_ref, "italic, yellow"),
                    Text("Running", style="bold green"),
                )
            else:
                table.add_row(
                    "Not While",
                    f,
                    Text("Ref: ").append(task.unique_ref, "italic yellow"),
                )

        return Panel(table, title=f"Plugin {self.name}")



"""The provided subclasses of a [`Node`][amltk.pipeline.node.Node]
that can be used can be assembled into a pipeline.
"""
from __future__ import annotations

import inspect
from collections.abc import Callable, Iterator, Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar, overload
from typing_extensions import Self, override

from more_itertools import all_unique, first_true

from amltk._functional import entity_name, mapping_select
from amltk.exceptions import (
    ComponentBuildError,
    DuplicateNamesError,
    NoChoiceMadeError,
    NodeNotFoundError,
)
from amltk.pipeline.node import Node, RichOptions
from amltk.randomness import randuid
from amltk.types import Config, Item, Space

if TYPE_CHECKING:
    from amltk.pipeline.node import NodeLike


T = TypeVar("T")
NodeT = TypeVar("NodeT", bound=Node)


@overload
def as_node(thing: Node, name: str | None = ...) -> Node:  # type: ignore
    ...


@overload
def as_node(thing: tuple[Node | NodeLike, ...], name: str | None = ...) -> Join:  # type: ignore
    ...


@overload
def as_node(thing: set[Node | NodeLike], name: str | None = ...) -> Choice:  # type: ignore
    ...


@overload
def as_node(thing: list[Node | NodeLike], name: str | None = ...) -> Sequential:  # type: ignore
    ...


@overload
def as_node(  # type: ignore
    thing: Callable[..., Item],
    name: str | None = ...,
) -> Component[Item, None]:
    ...


@overload
def as_node(thing: Item, name: str | None = ...) -> Fixed[Item]:
    ...


def as_node(  # noqa: PLR0911
    thing: Node | NodeLike[Item],
    name: str | None = None,
) -> Node | Choice | Join | Sequential | Fixed[Item]:
    """Convert a node, pipeline, set or tuple into a component, copying anything
    in the process and removing all linking to other nodes.

    Args:
        thing: The thing to convert
        name: The name of the node. If it already a node, it will be renamed to that
            one.

    Returns:
        The component
    """
    match thing:
        case set():
            return Choice(*thing, name=name)
        case tuple():
            return Join(*thing, name=name)
        case list():
            return Sequential(*thing, name=name)
        case Node():
            name = thing.name if name is None else name
            return thing.mutate(name=name)
        case type():
            return Component(thing, name=name)
        case thing if (inspect.isfunction(thing) or inspect.ismethod(thing)):
            return Component(thing, name=name)
        case _:
            return Fixed(thing, name=name)


@dataclass(init=False, frozen=True, eq=True)
class Component(Node[Item, Space]):
    """A [`Component`][amltk.pipeline.Component] of the pipeline with
    a possible item and **no children**.

    This is the basic building block of most pipelines, it accepts
    as it's [`item=`][amltk.pipeline.node.Node.item] some function that will be
    called with [`build_item()`][amltk.pipeline.components.Component.build_item] to
    build that one part of the pipeline.

    When [`build_item()`][amltk.pipeline.Component.build_item] is called, whatever
    the config of the component is at that time, will be used to construct the item.

    A common pattern is to use a [`Component`][amltk.pipeline.Component] to
    wrap a constructor, specifying the [`space=`][amltk.pipeline.node.Node.space]
    and [`config=`][amltk.pipeline.node.Node.config] to be used when building the
    item.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Component
    from sklearn.ensemble import RandomForestClassifier

    rf = Component(
        RandomForestClassifier,
        config={"max_depth": 3},
        space={"n_estimators": (10, 100)}
    )
    from amltk._doc import doc_print; doc_print(print, rf)  # markdown-exec: hide

    config = {"n_estimators": 50}  # Sample from some space or something
    configured_rf = rf.configure(config)

    estimator = configured_rf.build_item()
    from amltk._doc import doc_print; doc_print(print, estimator)  # markdown-exec: hide
    ```

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """

    item: Callable[..., Item]
    """A node which constructs an item in the pipeline."""

    nodes: tuple[()]
    """A component has no children."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="#E6AF2E")

    _NODES_INIT: ClassVar = None

    def __init__(
        self,
        item: Callable[..., Item],
        *,
        name: str | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """Initialize a component.

        Args:
            item: The item attached to this node.
            name: The name of the node. If not specified, the name will be
                generated from the item.
            config: The configuration for this node.
            space: The search space for this node. This will be used when
                [`search_space()`][amltk.pipeline.node.Node.search_space] is called.
            fidelities: The fidelities for this node.
            config_transform: A function that transforms the `config=` parameter
                during [`configure(config)`][amltk.pipeline.node.Node.configure]
                before return the new configured node. Useful for times where
                you need to combine multiple parameters into one.
            meta: Any meta information about this node.
        """
        super().__init__(
            name=name if name is not None else entity_name(item),
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )

    def build_item(self, **kwargs: Any) -> Item:
        """Build the item attached to this component.

        Args:
            **kwargs: Any additional arguments to pass to the item

        Returns:
            Item
                The built item
        """
        config = self.config or {}
        try:
            return self.item(**{**config, **kwargs})
        except TypeError as e:
            new_msg = f"Failed to build `{self.item=}` with `{self.config=}`.\n"
            if any(kwargs):
                new_msg += f"Extra {kwargs=} were also provided.\n"
            new_msg += (
                "If the item failed to initialize, a common reason can be forgetting"
                " to call `configure()` on the `Component` or the pipeline it is in or"
                " not calling `build()`/`build_item()` on the **returned** value of"
                " `configure()`.\n"
                "Reasons may also include not having fully specified the `config`"
                " initially, it having not being configured fully from `configure()`"
                " or from misspecfying parameters in the `space`."
            )
            raise ComponentBuildError(new_msg) from e


@dataclass(init=False, frozen=True, eq=True)
class Searchable(Node[None, Space]):  # type: ignore
    """A [`Searchable`][amltk.pipeline.Searchable]
    node of the pipeline which just represents a search space, no item attached.

    While not usually applicable to pipelines you want to build, this node
    is useful for creating a search space, especially if the real pipeline you
    want to optimize can not be built directly. For example, if you are optimize
    a script, you may wish to use a `Searchable` to represent the search space
    of that script.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Searchable

    script_space = Searchable({"mode": ["orange", "blue", "red"], "n": (10, 100)})
    from amltk._doc import doc_print; doc_print(print, script_space)  # markdown-exec: hide
    ```

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """  # noqa: E501

    item: None = None
    """A searchable has no item."""

    nodes: tuple[()] = ()
    """A searchable has no children."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="light_steel_blue")

    _NODES_INIT: ClassVar = None

    def __init__(
        self,
        space: Space | None = None,
        *,
        name: str | None = None,
        config: Config | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """Initialize a choice.

        Args:
            space: The search space for this node. This will be used when
                [`search_space()`][amltk.pipeline.node.Node.search_space] is called.
            name: The name of the node. If not specified, a random one will
                be generated.
            config: The configuration for this node. Useful for setting some
                default values.
            fidelities: The fidelities for this node.
            config_transform: A function that transforms the `config=` parameter
                during [`configure(config)`][amltk.pipeline.node.Node.configure]
                before return the new configured node. Useful for times where
                you need to combine multiple parameters into one.
            meta: Any meta information about this node.
        """
        if name is None:
            name = f"Searchable-{randuid(8)}"

        super().__init__(
            name=name,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )


@dataclass(init=False, frozen=True, eq=True)
class Fixed(Node[Item, None]):  # type: ignore
    """A [`Fixed`][amltk.pipeline.Fixed] part of the pipeline that
    represents something that can not be configured and used directly as is.

    It consists of an [`.item`][amltk.pipeline.node.Node.item] that is fixed,
    non-configurable and non-searchable. It also has no children.

    This is useful for representing parts of the pipeline that are fixed, for example
    if you have a pipeline that is a `Sequential` of nodes, but you want to
    fix the first component to be a `PCA` with `n_components=3`, you can use a `Fixed`
    to represent that.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Component, Fixed, Sequential
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.decomposition import PCA

    rf = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
    pca = Fixed(PCA(n_components=3))

    pipeline = Sequential(pca, rf, name="my_pipeline")
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide
    ```

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """

    item: Item
    """The fixed item that this node represents."""

    space: None = None
    """A fixed node has no search space."""

    fidelities: None = None
    """A fixed node has no search space."""

    config: None = None
    """A fixed node has no config."""

    config_transform: None = None
    """A fixed node has no config so no transform."""

    nodes: tuple[()] = ()
    """A fixed node has no children."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="#56351E")

    _NODES_INIT: ClassVar = None

    def __init__(  # noqa: D417
        self,
        item: Item,
        *,
        name: str | None = None,
        config: None = None,
        space: None = None,
        fidelities: None = None,
        config_transform: None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """Initialize a fixed node.

        Args:
            item: The item attached to this node. Will be fixed and can not
                be configured.
            name: The name of the node. If not specified, the name will be
                generated from the item.
            meta: Any meta information about this node.
        """
        super().__init__(
            name=name if name is not None else entity_name(item),
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )


@dataclass(init=False, frozen=True, eq=True)
class Sequential(Node[Item, Space]):
    """A [`Sequential`][amltk.pipeline.Sequential] set of operations in a pipeline.

    This indicates the different children in
    [`.nodes`][amltk.pipeline.Node.nodes] should act one after
    another, feeding the output of one into the next.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Component, Sequential
    from sklearn.decomposition import PCA
    from sklearn.ensemble import RandomForestClassifier

    pipeline = Sequential(
        PCA(n_components=3),
        Component(RandomForestClassifier, space={"n_estimators": (10, 100)}),
        name="my_pipeline"
    )
    from amltk._doc import doc_print; doc_print(print, pipeline)  # markdown-exec: hide
    ```

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """

    nodes: tuple[Node, ...]
    """The nodes ordered in series."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(
        panel_color="#7E6B8F",
        node_orientation="vertical",
    )
    _NODES_INIT: ClassVar = "args"

    def __init__(
        self,
        *nodes: Node | NodeLike,
        name: str | None = None,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """Initialize a sequential node.

        Args:
            nodes: The nodes that this node leads to. In the case of a `Sequential`,
                the order here matters and it signifies that data should first
                be passed through the first node, then the second, etc.
            item: The item attached to this node (if any).
            name: The name of the node. If not specified, the name will be
                randomly generated.
            config: The configuration for this node.
            space: The search space for this node. This will be used when
                [`search_space()`][amltk.pipeline.node.Node.search_space] is called.
            fidelities: The fidelities for this node.
            config_transform: A function that transforms the `config=` parameter
                during [`configure(config)`][amltk.pipeline.node.Node.configure]
                before return the new configured node. Useful for times where
                you need to combine multiple parameters into one.
            meta: Any meta information about this node.
        """
        _nodes = tuple(as_node(n) for n in nodes)

        # Perhaps we need to do a deeper check on this...
        if not all_unique(_nodes, key=lambda node: node.name):
            raise DuplicateNamesError(self)

        if name is None:
            name = f"Seq-{randuid(8)}"

        super().__init__(
            *_nodes,
            name=name,
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )

    @property
    def tail(self) -> Node:
        """The last step in the pipeline."""
        return self.nodes[-1]

    def __len__(self) -> int:
        """Get the number of nodes in the pipeline."""
        return len(self.nodes)

    @override
    def __rshift__(self, other: Node | NodeLike) -> Sequential:
        other_node = as_node(other)
        if any(other_node.name == node.name for node in self.nodes):
            raise ValueError(
                f"Can't handle node with name '{other_node.name} as"
                f" there is already a node called '{other_node.name}' in {self.name}",
            )

        nodes = (*tuple(as_node(n) for n in self.nodes), other_node)
        return self.mutate(name=self.name, nodes=nodes)

    @override
    def walk(
        self,
        path: Sequence[Node] | None = None,
    ) -> Iterator[tuple[list[Node], Node]]:
        """Walk the nodes in this chain.

        Args:
            path: The current path to this node

        Yields:
            The parents of the node and the node itself
        """
        path = list(path) if path is not None else []
        yield path, self

        path = [*path, self]
        for node in self.nodes:
            yield from node.walk(path=path)

            # Append the previous node so that the next node in the sequence is
            # lead to from the previous node
            path = [*path, node]


@dataclass(init=False, frozen=True, eq=True)
class Choice(Node[Item, Space]):
    """A [`Choice`][amltk.pipeline.Choice] between different subcomponents.

    This indicates that a choice should be made between the different children in
    [`.nodes`][amltk.pipeline.Node.nodes], usually done when you
    [`configure()`][amltk.pipeline.node.Node.configure] with some `config` from
    a [`search_space()`][amltk.pipeline.node.Node.search_space].

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Choice, Component
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.neural_network import MLPClassifier

    rf = Component(RandomForestClassifier, space={"n_estimators": (10, 100)})
    mlp = Component(MLPClassifier, space={"activation": ["logistic", "relu", "tanh"]})

    estimator_choice = Choice(rf, mlp, name="estimator")
    from amltk._doc import doc_print; doc_print(print, estimator_choice)  # markdown-exec: hide
    ```

    !!! warning "Order of nodes"

        The given nodes of a choice are always ordered according
        to their name, so indexing `choice.nodes` may not be reliable
        if modifying the choice dynamically.

        Please use `choice["name"]` to access the nodes instead.

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """  # noqa: E501

    nodes: tuple[Node, ...]
    """The choice of possible nodes that this choice could take."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="#FF4500")
    _NODES_INIT: ClassVar = "args"

    def __init__(
        self,
        *nodes: Node | NodeLike,
        name: str | None = None,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """Initialize a choice node.

        Args:
            nodes: The nodes that should be chosen between for this node.
            item: The item attached to this node (if any).
            name: The name of the node. If not specified, the name will be
                randomly generated.
            config: The configuration for this node.
            space: The search space for this node. This will be used when
                [`search_space()`][amltk.pipeline.node.Node.search_space] is called.
            fidelities: The fidelities for this node.
            config_transform: A function that transforms the `config=` parameter
                during [`configure(config)`][amltk.pipeline.node.Node.configure]
                before return the new configured node. Useful for times where
                you need to combine multiple parameters into one.
            meta: Any meta information about this node.
        """
        _nodes: tuple[Node, ...] = tuple(
            sorted((as_node(n) for n in nodes), key=lambda n: n.name),
        )
        if not all_unique(_nodes, key=lambda node: node.name):
            raise ValueError(
                f"Can't handle nodes as we can not generate a __choice__ for {nodes=}."
                "\nAll nodes must have a unique name. Please provide a `name=` to them",
            )

        if name is None:
            name = f"Choice-{randuid(8)}"

        super().__init__(
            *_nodes,
            name=name,
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )

    @override
    def __or__(self, other: Node | NodeLike) -> Choice:
        other_node = as_node(other)
        if any(other_node.name == node.name for node in self.nodes):
            raise ValueError(
                f"Can't handle node with name '{other_node.name} as"
                f" there is already a node called '{other_node.name}' in {self.name}",
            )

        nodes = tuple(
            sorted(
                [as_node(n) for n in self.nodes] + [other_node],
                key=lambda n: n.name,
            ),
        )
        return self.mutate(name=self.name, nodes=nodes)

    def chosen(self) -> Node:
        """The chosen branch.

        Returns:
            The chosen branch
        """
        match self.config:
            case {"__choice__": choice}:
                chosen = first_true(
                    self.nodes,
                    pred=lambda node: node.name == choice,
                    default=None,
                )
                if chosen is None:
                    raise NodeNotFoundError(choice, self.name)

                return chosen
            case _:
                raise NoChoiceMadeError(self.name)

    @override
    def configure(
        self,
        config: Config,
        *,
        prefixed_name: bool | None = None,
        transform_context: Any | None = None,
        params: Mapping[str, Any] | None = None,
    ) -> Self:
        """Configure this node and anything following it with the given config.

        !!! note "Configuring a choice"

            For a Choice, if the config has a `__choice__` key, then only the node
            chosen will be configured. The others will not be configured at all and
            their config will be discarded.

        Args:
            config: The configuration to apply
            prefixed_name: Whether items in the config are prefixed by the names
                of the nodes.
                * If `None`, the default, then `prefixed_name` will be assumed to
                    be `True` if this node has a next node or if the config has
                    keys that begin with this nodes name.
                * If `True`, then the config will be searched for items prefixed
                    by the name of the node (and subsequent chained nodes).
                * If `False`, then the config will be searched for items without
                    the prefix, i.e. the config keys are exactly those matching
                    this nodes search space.
            transform_context: Any context to give to `config_transform=` of individual
                nodes.
            params: The params to match any requests when configuring this node.
                These will match against any ParamRequests in the config and will
                be used to fill in any missing values.

        Returns:
            The configured node
        """
        # Get the config for this node
        match prefixed_name:
            case True:
                config = mapping_select(config, f"{self.name}:")
            case False:
                pass
            case None if any(k.startswith(f"{self.name}:") for k in config):
                config = mapping_select(config, f"{self.name}:")
            case None:
                pass

        _kwargs: dict[str, Any] = {}

        # Configure all the branches if exists
        # This part is what differs for a Choice
        if len(self.nodes) > 0:
            choice_made = config.get("__choice__", None)
            if choice_made is not None:
                matching_child = first_true(
                    self.nodes,
                    pred=lambda node: node.name == choice_made,
                    default=None,
                )
                if matching_child is None:
                    raise ValueError(
                        f"Can not find matching child for choice {self.name} with child"
                        f" {choice_made}."
                        "\nPlease check the config and ensure that the choice is one of"
                        f" {[n.name for n in self.nodes]}."
                        f"\nThe config recieved at this choice node was {config=}.",
                    )

                # We still iterate over all of them just to ensure correct ordering
                nodes = tuple(
                    node.copy()
                    if node.name != choice_made
                    else matching_child.configure(
                        config,
                        prefixed_name=True,
                        transform_context=transform_context,
                        params=params,
                    )
                    for node in self.nodes
                )
                _kwargs["nodes"] = nodes
            else:
                nodes = tuple(
                    node.configure(
                        config,
                        prefixed_name=True,
                        transform_context=transform_context,
                        params=params,
                    )
                    for node in self.nodes
                )
                _kwargs["nodes"] = nodes

        this_config = {
            hp: v
            for hp, v in config.items()
            if (
                ":" not in hp
                and not any(hp.startswith(f"{node.name}") for node in self.nodes)
            )
        }
        if self.config is not None:
            this_config = {**self.config, **this_config}

        this_config = dict(self._fufill_param_requests(this_config, params=params))

        if self.config_transform is not None:
            this_config = dict(self.config_transform(this_config, transform_context))

        if len(this_config) > 0:
            _kwargs["config"] = dict(this_config)

        return self.mutate(**_kwargs)


@dataclass(init=False, frozen=True, eq=True)
class Split(Node[Item, Space]):
    """A [`Split`][amltk.pipeline.Split] of data in a pipeline.

    This indicates the different children in
    [`.nodes`][amltk.pipeline.Node.nodes] should
    act in parallel but on different subsets of data.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Component, Split
    from sklearn.impute import SimpleImputer
    from sklearn.preprocessing import OneHotEncoder
    from sklearn.compose import make_column_selector

    categorical_pipeline = [
        SimpleImputer(strategy="constant", fill_value="missing"),
        OneHotEncoder(drop="first"),
    ]
    numerical_pipeline = Component(SimpleImputer, space={"strategy": ["mean", "median"]})

    preprocessor = Split(
        {
            "categories": categorical_pipeline,
            "numerical": numerical_pipeline,
        },
        config={
            "categories": make_column_selector(dtype_include="category"),
            "numerical": make_column_selector(dtype_exclude="category"),
        },
        name="my_split"
    )
    from amltk._doc import doc_print; doc_print(print, preprocessor)  # markdown-exec: hide
    ```

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """  # noqa: E501

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(
        panel_color="#777DA7",
        node_orientation="horizontal",
    )

    _NODES_INIT: ClassVar = "args"

    def __init__(
        self,
        *nodes: Node | NodeLike | dict[str, Node | NodeLike],
        name: str | None = None,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """Initialize a split node.

        Args:
            nodes: The nodes that this node leads to. You may also provide
                a dictionary where the keys are the names of the nodes and
                the values are the nodes or list of nodes themselves.
            item: The item attached to this node. The object created by `item`
                should be capable of figuring out how to deal with its child nodes.
            name: The name of the node. If not specified, the name will be
                generated from the item.
            config: The configuration for this split.
            space: The search space for this node. This will be used when
                [`search_space()`][amltk.pipeline.node.Node.search_space] is called.
            fidelities: The fidelities for this node.
            config_transform: A function that transforms the `config=` parameter
                during [`configure(config)`][amltk.pipeline.node.Node.configure]
                before return the new configured node. Useful for times where
                you need to combine multiple parameters into one.
            meta: Any meta information about this node.
        """
        if any(isinstance(n, dict) for n in nodes):
            if len(nodes) > 1:
                raise ValueError(
                    "Can't handle multiple nodes with a dictionary as a node.\n"
                    f"{nodes=}",
                )
            _node = nodes[0]
            assert isinstance(_node, dict)

            def _construct(key: str, value: Node | NodeLike) -> Node:
                match value:
                    case list():
                        return Sequential(*value, name=key)
                    case set() | tuple():
                        return as_node(value, name=key)
                    case _:
                        return Sequential(value, name=key)

            _nodes = tuple(_construct(key, value) for key, value in _node.items())
        else:
            _nodes = tuple(as_node(n) for n in nodes)

        if not all_unique(_nodes, key=lambda node: node.name):
            raise ValueError(
                f"Can't handle nodes they do not all contain unique names, {nodes=}."
                "\nAll nodes must have a unique name. Please provide a `name=` to them",
            )

        if name is None:
            name = f"Split-{randuid(8)}"

        super().__init__(
            *_nodes,
            name=name,
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )


@dataclass(init=False, frozen=True, eq=True)
class Join(Node[Item, Space]):
    """[`Join`][amltk.pipeline.Join] together different parts of the pipeline.

    This indicates the different children in
    [`.nodes`][amltk.pipeline.Node.nodes] should act in tandem with one
    another, for example, concatenating the outputs of the various members of the
    `Join`.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Join, Component
    from sklearn.decomposition import PCA
    from sklearn.feature_selection import SelectKBest

    pca = Component(PCA, space={"n_components": (1, 3)})
    kbest = Component(SelectKBest, space={"k": (1, 3)})

    join = Join(pca, kbest, name="my_feature_union")
    from amltk._doc import doc_print; doc_print(print, join)  # markdown-exec: hide
    ```

    See Also:
        * [`Node`][amltk.pipeline.node.Node]
    """

    nodes: tuple[Node, ...]
    """The nodes that should be joined together in parallel."""

    RICH_OPTIONS: ClassVar[RichOptions] = RichOptions(panel_color="#7E6B8F")
    _NODES_INIT: ClassVar = "args"

    def __init__(
        self,
        *nodes: Node | NodeLike,
        name: str | None = None,
        item: Item | Callable[[Item], Item] | None = None,
        config: Config | None = None,
        space: Space | None = None,
        fidelities: Mapping[str, Any] | None = None,
        config_transform: Callable[[Config, Any], Config] | None = None,
        meta: Mapping[str, Any] | None = None,
    ):
        """Initialize a join node.

        Args:
            nodes: The nodes that should be joined together in parallel.
            item: The item attached to this node (if any).
            name: The name of the node. If not specified, the name will be
                randomly generated.
            config: The configuration for this node.
            space: The search space for this node. This will be used when
                [`search_space()`][amltk.pipeline.node.Node.search_space] is called.
            fidelities: The fidelities for this node.
            config_transform: A function that transforms the `config=` parameter
                during [`configure(config)`][amltk.pipeline.node.Node.configure]
                before return the new configured node. Useful for times where
                you need to combine multiple parameters into one.
            meta: Any meta information about this node.
        """
        _nodes = tuple(as_node(n) for n in nodes)
        if not all_unique(_nodes, key=lambda node: node.name):
            raise ValueError(
                f"Can't handle nodes they do not all contain unique names, {nodes=}."
                "\nAll nodes must have a unique name. Please provide a `name=` to them",
            )

        if name is None:
            name = f"Join-{randuid(8)}"

        super().__init__(
            *_nodes,
            name=name,
            item=item,
            config=config,
            space=space,
            fidelities=fidelities,
            config_transform=config_transform,
            meta=meta,
        )

    @override
    def __and__(self, other: Node | NodeLike) -> Join:
        other_node = as_node(other)
        if any(other_node.name == node.name for node in self.nodes):
            raise ValueError(
                f"Can't handle node with name '{other_node.name} as"
                f" there is already a node called '{other_node.name}' in {self.name}",
            )

        nodes = (*tuple(as_node(n) for n in self.nodes), other_node)
        return self.mutate(name=self.name, nodes=nodes)



"""The
[`ThreadPoolCTLPlugin`][amltk.scheduling.plugins.threadpoolctl.ThreadPoolCTLPlugin]
if useful for parallel training of models. Without limiting with
threadpoolctl, the number of threads used by a given model may
oversubscribe to resources and cause significant slowdowns.

This is the mechanism employed by scikit-learn to limit the number of
threads used by a given model.

See [threadpoolctl documentation](https://github.com/joblib/threadpoolctl).

!!! tip "Requirements"

    This requires `threadpoolctl` which can be installed with:

    ```bash
    pip install amltk[threadpoolctl]

    # Or directly
    pip install threadpoolctl
    ```

??? tip "Usage"

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins.threadpoolctl import ThreadPoolCTLPlugin

    scheduler = Scheduler.with_processes(1)

    def f() -> None:
        # ... some task that respects the limits set by threadpoolctl
        pass

    task = scheduler.task(f, plugins=ThreadPoolCTLPlugin(max_threads=1))
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
"""
from __future__ import annotations

import logging
from collections.abc import Callable
from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar
from typing_extensions import ParamSpec, Self, override

from amltk.scheduling.plugins.plugin import Plugin

if TYPE_CHECKING:
    from rich.panel import Panel

    from amltk.scheduling.task import Task

P = ParamSpec("P")
R = TypeVar("R")

logger = logging.getLogger(__name__)


class _ThreadPoolLimiter(Generic[P, R]):
    def __init__(
        self,
        fn: Callable[P, R],
        max_threads: int | dict[str, int] | None = None,
        user_api: str | None = None,
    ):
        super().__init__()
        self.fn = fn
        self.max_threads = max_threads
        self.user_api = user_api

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        import threadpoolctl

        controller = threadpoolctl.ThreadpoolController()
        with controller.limit(limits=self.max_threads, user_api=self.user_api):
            logger.debug(f"threadpoolctl: {controller.info()}")
            return self.fn(*args, **kwargs)


class ThreadPoolCTLPlugin(Plugin):
    """A plugin that limits the usage of threads in a task.

    This plugin is used to make utilize threadpoolctl with tasks,
    useful for parallel training of models. Without limiting with
    threadpoolctl, the number of threads used by a given model may
    oversubscribe to resources and cause significant slowdowns.

    Attributes:
        max_calls: The maximum number of calls to the task.
        max_concurrent: The maximum number of calls of this task that can
            be in the queue.
    """

    name: ClassVar = "threadpoolctl-plugin"
    """The name of the plugin."""

    @override
    def __init__(
        self,
        max_threads: int | dict[str, int] | None = None,
        user_api: str | None = None,
    ):
        """Initialize the plugin.

        Args:
            max_threads: The maximum number of threads to use.
            user_api: The user API to limit.
        """
        super().__init__()
        self.max_threads = max_threads
        self.user_api = user_api
        self.task: Task | None = None

    @override
    def attach_task(self, task: Task) -> None:
        """Attach the plugin to a task."""
        self.task = task

    @override
    def pre_submit(
        self,
        fn: Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> tuple[Callable[P, R], tuple, dict] | None:
        """Pre-submit hook.

        Wrap the function in something that will activate threadpoolctl
        when the function is called.
        """
        assert self.task is not None

        fn = _ThreadPoolLimiter(
            fn=fn,
            max_threads=self.max_threads,
            user_api=self.user_api,
        )
        return fn, args, kwargs

    @override
    def copy(self) -> Self:
        """Return a copy of the plugin.

        Please see [`Plugin.copy()`][amltk.Plugin.copy].
        """
        return self.__class__(max_threads=self.max_threads, user_api=self.user_api)

    @override
    def __rich__(self) -> Panel:
        from rich.panel import Panel
        from rich.pretty import Pretty
        from rich.table import Table

        table = Table(
            "Max Threads",
            "User-API",
            padding=(0, 1),
            show_edge=False,
            box=None,
        )
        table.add_row(Pretty(self.max_threads), Pretty(self.user_api))
        return Panel(table, title=f"Plugin {self.name}")



"""Utilities for dealing with randomness."""
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from amltk.types import Seed

MAX_INT = np.iinfo(np.int32).max
ALPHABET = list("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789")


def as_rng(seed: Seed | None = None) -> np.random.Generator:
    """Converts a valid seed arg into a numpy.random.Generator instance.

    Args:
        seed: The seed to use

    Returns:
        A valid np.random.Generator object to use
    """
    match seed:
        case None | int() | np.integer():
            return np.random.default_rng(seed)
        case np.random.Generator():
            return seed
        case np.random.RandomState():
            _seed = seed.randint(0, MAX_INT)
            return np.random.default_rng(_seed)

    raise ValueError(f"Can't {seed=} ({type(seed)}) to create numpy.random.Generator")


def as_randomstate(seed: Seed | None = None) -> np.random.RandomState:
    """Converts a valid seed arg into a numpy.random.RandomState instance.

    Args:
        seed: The seed to use

    Returns:
        A valid np.random.RandomSTate object to use
    """
    match seed:
        case None | int() | np.integer():
            return np.random.RandomState(seed)
        case np.random.RandomState():
            return seed
        case np.random.Generator():
            _seed = seed.integers(0, MAX_INT)
            return np.random.RandomState(_seed)

    raise ValueError(f"Can't {seed=} ({type(seed)}) to create numpy.random.RandomState")


def as_int(seed: Seed | None = None) -> int:
    """Converts a valid seed arg into an integer.

    Args:
        seed: The seed to use

    Returns:
        A valid integer to use as a seed
    """
    match seed:
        case None:
            return int(np.random.default_rng().integers(0, MAX_INT))
        case np.integer() | int():
            return int(seed)
        case np.random.Generator():
            return int(seed.integers(0, MAX_INT))
        case np.random.RandomState():
            return int(seed.randint(0, MAX_INT))

    raise ValueError(f"Can't {seed=} ({type(seed)}) to create int")


def randuid(
    k: int = 8,
    *,
    charset: Sequence[str] = ALPHABET,
    seed: Seed | None = None,
) -> str:
    """Generate a random alpha-numeric uuid of a specified length.

    See: https://stackoverflow.com/a/56398787/5332072

    Args:
        k: The length of the uuid to generate
        charset: The charset to use
        seed: The seed to use

    Returns:
        A random uid
    """
    rng = as_rng(seed)
    return "".join(rng.choice(np.asarray(charset), size=k))



"""The [`History`][amltk.optimization.History] is
used to keep a structured record of what occured with
[`Trial`][amltk.optimization.Trial]s and their associated
[`Report`][amltk.optimization.Trial.Report]s.

??? tip "Usage"

    ```python exec="true" source="material-block" html="true" hl_lines="19 23-24"
    from amltk.optimization import Trial, History, Metric
    from amltk.store import PathBucket

    loss = Metric("loss", minimize=True)

    def target_function(trial: Trial) -> Trial.Report:
        x = trial.config["x"]
        y = trial.config["y"]
        trial.store({"config.json": trial.config})

        loss = x**2 - y
        return trial.success(loss=loss)

    # ... usually obtained from an optimizer
    bucket = PathBucket("all-trial-results")
    history = History()

    for x, y in zip([1, 2, 3], [4, 5, 6]):
        trial = Trial(name="some-unique-name", config={"x": x, "y": y}, bucket=bucket, metrics=[loss])
        report = target_function(trial)
        history.add(report)

    print(history.df())
    bucket.rmdir()  # markdon-exec: hide
    ```

You'll often need to perform some operations on a
[`History`][amltk.optimization.History] so we provide some utility functions here:

* [`filter(key=...)`][amltk.optimization.History.filter] - Filters the history by some
    predicate, e.g. `#!python history.filter(lambda report: report.status == "success")`
* [`groupby(key=...)`][amltk.optimization.History.groupby] - Groups the history by some
    key, e.g. `#!python history.groupby(lambda report: report.config["x"] < 5)`
* [`sortby(key=...)`][amltk.optimization.History.sortby] - Sorts the history by some
    key, e.g. `#!python history.sortby(lambda report: report.profiles["trial"].time.end)`

There is also some serialization capabilities built in, to allow you to store
your reports and load them back in later:

* [`df(...)`][amltk.optimization.History.df] - Output a `pd.DataFrame` of all
 the information available.
* [`from_df(...)`][amltk.optimization.History.from_df] - Create a `History` from
    a `pd.DataFrame`.

You can also retrieve individual reports from the history by using their
name, e.g. `#!python history["some-unique-name"]` or iterate through
the history with `#!python for report in history: ...`.
"""  # noqa: E501
from __future__ import annotations

import operator
from collections import defaultdict
from collections.abc import Callable, Hashable, Iterable, Iterator
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, TypeVar, overload
from typing_extensions import override

import pandas as pd

from amltk._functional import compare_accumulate
from amltk._richutil import RichRenderable
from amltk.optimization.trial import Trial
from amltk.types import Comparable

if TYPE_CHECKING:
    from rich.console import RenderableType

    from amltk.optimization.metric import Metric

T = TypeVar("T")
CT = TypeVar("CT", bound=Comparable)
HashableT = TypeVar("HashableT", bound=Hashable)

# TODO: It might be faster to basically have the history
# always be a data frame, however this makes some things
# such as metrics a bit more difficult to work with.


@dataclass
class History(RichRenderable):
    """A history of trials.

    This is a collections of reports from trials, where you can access
    the reports by their trial name. It is unsorted in general, but
    by using [`sortby()`][amltk.optimization.History.sortby] you
    can sort the history.

    ```python exec="true" source="material-block" result="python" title="History"
    from amltk.optimization import Trial, History, Metric

    metric = Metric("cost", minimize=True)
    trials = [
        Trial(name=f"trial_{i}", config={"x": i}, metrics=[metric])
        for i in range(10)
    ]
    history = History()

    for trial in trials:
        x = trial.config["x"]
        report = trial.success(cost=x**2 - x*2 + 4)
        history.add(report)

    for report in history:
        print(f"{report.name=}, {report}")

    print(history.metrics)
    print(history.df())

    print(history.best())
    ```

    Attributes:
        reports: A mapping of trial names to reports.
    """

    reports: list[Trial.Report] = field(default_factory=list)
    metrics: dict[str, Metric] = field(default_factory=dict, repr=False)
    _lookup: dict[str, int] = field(default_factory=dict, repr=False)

    @classmethod
    def from_reports(cls, reports: Iterable[Trial.Report]) -> History:
        """Creates a history from reports.

        Args:
            reports: An iterable of reports.

        Returns:
            A history.
        """
        history = cls()
        history.add(reports)
        return history

    def best(self, metric: str | None = None) -> Trial.Report:
        """Returns the best report in the history.

        Args:
            metric: The metric to sort by. If `None`, it will use the
                first metric in the history. If there are multiple metrics
                and non are specified, it will raise an error.

        Returns:
            The best report.
        """
        if metric is None:
            if len(self.metrics) > 1:
                raise ValueError(
                    "There are multiple metrics in the history, "
                    "please specify which metric to sort by.",
                )

            _metric_def = next(iter(self.metrics.values()))
            _metric_name = _metric_def.name
        else:
            if metric not in self.metrics:
                raise ValueError(
                    f"Metric {metric} not found in history. "
                    f"Available metrics: {list(self.metrics.keys())}",
                )
            _metric_def = self.metrics[metric]
            _metric_name = metric

        _by = min if _metric_def.minimize else max
        return _by(self.reports, key=lambda r: r.metrics[_metric_name])

    def add(self, report: Trial.Report | Iterable[Trial.Report]) -> None:
        """Adds a report or reports to the history.

        Args:
            report: A report or reports to add.
        """
        match report:
            case Trial.Report():
                for m in report.metric_values:
                    if (_m := self.metrics.get(m.name)) is not None:
                        if m.metric != _m:
                            raise ValueError(
                                f"Metric {m.name} has conflicting definitions:"
                                f"\n{m.metric} != {_m}",
                            )
                    else:
                        self.metrics[m.name] = m.metric

                self.reports.append(report)
                self._lookup[report.name] = len(self.reports) - 1
            case Iterable():
                for _report in report:
                    # Could endless loop
                    if not isinstance(_report, Trial.Report):
                        raise TypeError(
                            f"Expected an iterable of Trial.Report"
                            f" but got {type(_report)} ({_report})"
                            f" in add({report=})",
                        )
                    self.add(_report)
            case _:
                raise TypeError(
                    f"Expected a Trial.Report or an iterable of Trial.Report"
                    f" but got {type(report)} ({report})",
                )

    def find(self, name: str) -> Trial.Report:
        """Finds a report by trial name.

        Args:
            name: The name of the trial.

        Returns:
            The report.
        """
        return self.reports[self._lookup[name]]

    def df(
        self,
        *,
        profiles: bool = True,
        configs: bool = True,
        summary: bool = True,
        metrics: bool = True,
        normalize_time: bool | float = True,
    ) -> pd.DataFrame:
        """Returns a pandas DataFrame of the history.

        Each individual trial will be a row in the dataframe.

        !!! note "Prefixes"

            * `summary`: Entries will be prefixed with `#!python "summary:"`
            * `config`: Entries will be prefixed with `#!python "config:"`
            * `metrics`: Entries will be prefixed with `#!python "metrics:"`

        ```python exec="true" source="material-block" result="python" title="df" hl_lines="12"
        from amltk.optimization import Trial, History, Metric

        metric = Metric("cost", minimize=True)
        trials = [Trial(name=f"trial_{i}", config={"x": i}, metrics=[metric]) for i in range(10)]
        history = History()

        for trial in trials:
            x = trial.config["x"]
            report = trial.success(cost=x**2 - x*2 + 4)
            history.add(report)

        print(history.df())
        ```

        Args:
            profiles: Whether to include the profiles.
            configs: Whether to include the configs.
            summary: Whether to include the summary.
            metrics: Whether to include the metrics.
            normalize_time: Whether to normalize the time to the first
                report. If given a `#!python float`, it will normalize
                to that value.

                Will normalize all columns with `#!python "time:end"`.
                and `#!python "time:start"` in their name. It will use
                the time of the earliest report as the offset.

        Returns:
            A pandas DataFrame of the history.
        """  # noqa: E501
        if len(self) == 0:
            return pd.DataFrame()

        _df = pd.concat(
            [
                report.df(
                    profiles=profiles,
                    configs=configs,
                    summary=summary,
                    metrics=metrics,
                )
                for report in self.reports
            ],
        )
        _df = _df.convert_dtypes()

        match normalize_time:
            case True if "time:start" in _df.columns:
                time_columns = ("time:start", "time:end")
                cols = [c for c in _df.columns if c.endswith(time_columns)]
                _df[cols] -= _df["time:start"].min()
            case float():
                time_columns = ("time:start", "time:end")
                cols = [c for c in _df.columns if c.endswith(time_columns)]
                _df[cols] -= normalize_time
            case _:
                pass

        return _df

    def filter(self, key: Callable[[Trial.Report], bool]) -> History:
        """Filters the history by a predicate.

        ```python exec="true" source="material-block" result="python" title="filter" hl_lines="12"
        from amltk.optimization import Trial, History, Metric

        metric = Metric("cost", minimize=True)
        trials = [Trial(name=f"trial_{i}", config={"x": i}, metrics=[metric]) for i in range(10)]
        history = History()

        for trial in trials:
            x = trial.config["x"]
            report = trial.success(cost=x**2 - x*2 + 4)
            history.add(report)

        filtered_history = history.filter(lambda report: report.metrics["cost"] < 10)
        for report in filtered_history:
            cost = report.metrics["cost"]
            print(f"{report.name}, {cost=}, {report}")
        ```

        Args:
            key: A predicate to filter by.

        Returns:
            A new history with the filtered reports.
        """  # noqa: E501
        return History.from_reports([report for report in self.reports if key(report)])

    def groupby(
        self,
        key: Literal["status"] | Callable[[Trial.Report], Hashable],
    ) -> dict[Hashable, History]:
        """Groups the history by the values of a key.

        ```python exec="true" source="material-block" result="python" title="groupby" hl_lines="15"
        from amltk.optimization import Trial, History, Metric

        metric = Metric("cost", minimize=True)
        trials = [Trial(name=f"trial_{i}", config={"x": i}, metrics=[metric]) for i in range(10)]
        history = History()

        for trial in trials:
            x = trial.config["x"]
            if x % 2 == 0:
                report = trial.fail(cost=1_000)
            else:
                report = trial.success(cost=x**2 - x*2 + 4)
            history.add(report)

        for status, history in history.groupby("status").items():
            print(f"{status=}, {len(history)=}")
        ```

        You can pass a `#!python Callable` to group by any key you like:

        ```python exec="true" source="material-block" result="python"
        from amltk.optimization import Trial, History, Metric

        metric = Metric("cost", minimize=True)
        trials = [Trial(name=f"trial_{i}", config={"x": i}, metrics=[metric]) for i in range(10)]
        history = History()

        for trial in trials:
            x = trial.config["x"]
            report = trial.fail(cost=x)
            history.add(report)

        for below_5, history in history.groupby(lambda r: r.metrics["cost"] < 5).items():
            print(f"{below_5=}, {len(history)=}")
        ```

        Args:
            key: A key to group by. If `"status"` is passed, the history will be
                grouped by the status of the reports.

        Returns:
            A mapping of keys to histories.
        """  # noqa: E501
        d = defaultdict(list)

        if key == "status":
            key = operator.attrgetter("status")

        for report in self.reports:
            d[key(report)].append(report)

        return {k: History.from_reports(v) for k, v in d.items()}

    def incumbents(
        self,
        key: Callable[[Trial.Report, Trial.Report], bool] | str,
        *,
        sortby: Callable[[Trial.Report], Comparable] | str = lambda r: r.reported_at,
        reverse: bool | None = None,
        ffill: bool = False,
    ) -> list[Trial.Report]:
        """Returns a trace of the incumbents, where only the report that is better than the previous
        best report is kept.

        ```python exec="true" source="material-block" result="python" title="incumbents"
        from amltk.optimization import Trial, History, Metric

        metric = Metric("cost", minimize=True)
        trials = [Trial(name=f"trial_{i}", config={"x": i}, metrics=[metric]) for i in range(10)]
        history = History()

        for trial in trials:
            x = trial.config["x"]
            report = trial.success(cost=x**2 - x*2 + 4)
            history.add(report)

        incumbents = (
            history
            .incumbents("cost", sortby=lambda r: r.reported_at)
        )
        for report in incumbents:
            print(f"{report.metrics=}, {report.config=}")
        ```

        Args:
            key: The key to use. If given a str, it will use that as the
                key to use in the metrics, defining if one report is better
                than another. If given a `#!python Callable`, it should
                return a `bool`, indicating if the first argument report
                is better than the second argument report.
            sortby: The key to sort by. If given a str, it will sort by
                the value of that key in the `.metrics` and also filter
                out anything that does not contain this key.
                By default, it will sort by the end time of the report.
            reverse: Whether to sort in some given order. By
                default (`None`), if given a metric key, the reports with
                the best metric values will be sorted first. If
                given a `#!python Callable`, the reports with the
                smallest values will be sorted first. Using
                `reverse=True` will always reverse this order, while
                `reverse=False` will always preserve it.
            ffill: Whether to forward fill the incumbents. This means that
                if a report is not an incumbent, it will be replaced with
                the current best. This is useful if you want to
                visualize the incumbents over some x axis, where the
                you have a point at every place along the axis.

        Returns:
            The history of incumbents.
        """  # noqa: E501
        match key:
            case str():
                metric = self.metrics[key]
                __op = operator.lt if metric.minimize else operator.gt  # type: ignore
                op = lambda r1, r2: __op(r1.metrics[key], r2.metrics[key])
            case _:
                op = key

        sorted_reports = self.sortby(sortby, reverse=reverse)
        return list(compare_accumulate(sorted_reports, op=op, ffill=ffill))

    def sortby(
        self,
        key: Callable[[Trial.Report], Comparable] | str,
        *,
        reverse: bool | None = None,
    ) -> list[Trial.Report]:
        """Sorts the history by a key and returns a sorted History.

        ```python exec="true" source="material-block" result="python" title="sortby" hl_lines="15"
        from amltk.optimization import Trial, History, Metric

        metric = Metric("cost", minimize=True)
        trials = [Trial(name=f"trial_{i}", config={"x": i}, metrics=[metric]) for i in range(10)]
        history = History()

        for trial in trials:
            x = trial.config["x"]
            report = trial.success(cost=x**2 - x*2 + 4)
            history.add(report)

        trace = (
            history
            .filter(lambda report: report.status == "success")
            .sortby("cost")
        )

        for report in trace:
            print(f"{report.metrics}, {report}")
        ```

        Args:
            key: The key to sort by. If given a str, it will sort by
                the value of that key in the `.metrics` and also filter
                out anything that does not contain this key.
            reverse: Whether to sort in some given order. By
                default (`None`), if given a metric key, the reports with
                the best metric values will be sorted first. If
                given a `#!python Callable`, the reports with the
                smallest values will be sorted first. Using
                `reverse=True` will always reverse this order, while
                `reverse=False` will always preserve it.

        Returns:
            A sorted list of reports
        """  # noqa: E501
        # If given a str, filter out anything that doesn't have that key
        if isinstance(key, str):
            history = self.filter(lambda report: key in report.metric_names)
            sort_key: Callable[[Trial.Report], Comparable] = lambda r: r.metrics[key]
            reverse = (
                reverse if reverse is not None else (not self.metrics[key].minimize)
            )
        else:
            history = self
            sort_key = key
            reverse = False if reverse is None else reverse

        return sorted(history.reports, key=sort_key, reverse=reverse)

    @overload
    def __getitem__(self, key: int | str) -> Trial.Report:
        ...

    @overload
    def __getitem__(self, key: slice) -> Trial.Report:
        ...

    def __getitem__(  # type: ignore
        self,
        key: int | str | slice,
    ) -> Trial.Report | History:
        if isinstance(key, str):
            return self.find(key)
        if isinstance(key, int):
            return self.reports[key]

        return History.from_reports(self.reports[key])

    def __iter__(self) -> Iterator[Trial.Report]:
        return iter(self.reports)

    def __len__(self) -> int:
        return len(self.reports)

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, History):
            return NotImplemented
        return (
            self.reports == other.reports
            and self.metrics == other.metrics
            and self._lookup == other._lookup
        )

    @classmethod
    def from_df(cls, df: pd.DataFrame) -> History:
        """Loads a history from a pandas DataFrame.

        Args:
            df: The DataFrame to load the history from.

        Returns:
            A History.
        """
        if len(df) == 0:
            return cls()
        return History.from_reports(Trial.Report.from_df(s) for _, s in df.iterrows())

    @override
    def __rich__(self) -> RenderableType:
        from amltk._richutil import df_to_table

        return df_to_table(self.df(profiles=False), title="History", expand=True)

    @override
    def _repr_html_(self) -> str:
        return str(self.df().to_html())



"""Evaluation protocols for how a trial and a pipeline should be evaluated.

TODO: Sorry
"""
from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING

from amltk.scheduling import Plugin

if TYPE_CHECKING:
    from amltk.optimization import Trial
    from amltk.pipeline import Node
    from amltk.scheduling import Scheduler, Task


class EvaluationProtocol:
    """A protocol for how a trial should be evaluated on a pipeline."""

    fn: Callable[[Trial, Node], Trial.Report]

    def task(
        self,
        scheduler: Scheduler,
        plugins: Plugin | Iterable[Plugin] | None = None,
    ) -> Task[[Trial, Node], Trial.Report]:
        """Create a task for this protocol.

        Args:
            scheduler: The scheduler to use for the task.
            plugins: The plugins to use for the task.

        Returns:
            The created task.
        """
        _plugins: tuple[Plugin, ...]
        match plugins:
            case None:
                _plugins = ()
            case Plugin():
                _plugins = (plugins,)
            case Iterable():
                _plugins = tuple(plugins)

        return scheduler.task(self.fn, plugins=_plugins)


class CustomProtocol(EvaluationProtocol):
    """A custom evaluation protocol based on a user function."""

    def __init__(self, fn: Callable[[Trial, Node], Trial.Report]) -> None:
        """Initialize the protocol.

        Args:
            fn: The function to use for the evaluation.
        """
        super().__init__()
        self.fn = fn



"""Operations on pipelines."""
from __future__ import annotations

import sys
from collections.abc import Callable, Iterator
from functools import partial
from itertools import product
from typing import TypeVar

from amltk.pipeline.components import Choice
from amltk.pipeline.node import Node

MAX_INT = sys.maxsize
NodeT1 = TypeVar("NodeT1", bound=Node)
NodeT2 = TypeVar("NodeT2", bound=Node)


def factorize(
    node: NodeT1,
    *,
    min_depth: int = 0,
    max_depth: int | None = None,
    current_depth: int = 0,
    factor_by: Callable[[Node], bool] | None = None,
    assign_child: Callable[[NodeT2, Node], NodeT2] | None = None,
) -> Iterator[NodeT1]:
    """Factorize a pipeline into all possibilities of its children.

    When dealing with a large pipeline with many choices at various levels,
    it can be useful to factorize the pipeline into all possible pipelines.
    This effectively returns a new pipeline for every possible choice in the
    pipeline.

    ```python exec="true" source="material-block" html="true"
    from amltk.pipeline import Sequential, Choice, Node, factorize

    pipeline = Sequential(
        Choice(Node(name="hi"), Node(name="hello"), name="choice"),
        Node(name="banana"),
        name="pipeline",
    )

    from amltk._doc import doc_print; _print = print; print = lambda thing: doc_print(_print, thing)  # markdown-exec: hide
    print(pipeline)
    for i, possibility in enumerate(factorize(pipeline)):
        print(f"Pipeline {i}:")
        print(possibility)
    ```

    Args:
        node: The node to factorize.
        min_depth: The minimum depth at which to factorize. If the node is at a
            depth less than this, it will not be factorized.
            Depth is calculated as the node distance from node which is passed in
            plus the `current_depth`.
        max_depth: The maximum depth at which to factorize. If the node is at a
            depth greater than this, it will not be factorized.
            Depth is calculated as the node distance from node which is passed in
            plus the `current_depth`. If `None`, there is no maximum depth.
        current_depth: The current depth of the node. This is used internally but
            can also be used externally to factorize a sub-pipeline.
        factor_by: A function that takes a node and returns True if it
            should be factorized into its children, False otherwise. By
            default, it will split only Choice nodes. One useful example
            is to only factor on a particular name of a node.

            ```python
            pipeline_per_estimator = list(factorize(
                pipeline,
                factor_by=lambda _node: _node.name == "estimator"
            ))
            ```

        assign_child: A function that takes a node and a child and
            returns a new node with that child assigned to it. By default,
            it will mutate the node so that it has that child as its
            only child. You may wish to pass in custom functionality if there
            is more than one way to assign a child to a node or extra logic must
            be done to the nodes properties.

            It should return the same type of node as the one passed in.

    Returns:
        An iterator over all possible pipelines.
    """  # noqa: E501
    if max_depth is None:
        max_depth = MAX_INT

    if current_depth < 0:
        raise ValueError("current_depth cannot be less than 0")

    # We can exit early as there's no chance we factorize past this point
    if current_depth > max_depth:
        yield node.copy()
        return

    # NOTE: These two functions below are defined here instead to allow custom
    # Node types in the future. The default behaviour is defined to just split
    # Choice nodes and assign a child to one is to just mutate the node so
    # that it has that child as its only child.
    if factor_by is None:
        factor_by = lambda _node: isinstance(_node, Choice)

    if assign_child is None:
        assign_child = lambda _node, _child: _node.mutate(nodes=(_child,))

    _factorize = partial(
        factorize,
        factor_by=factor_by,
        assign_child=assign_child,
        current_depth=current_depth + 1,
        min_depth=min_depth,
        max_depth=max_depth,
    )

    match node:
        case Node(nodes=()):
            # Base case, there's no further possibility to factorize
            yield node.copy()
        case Node(nodes=children) if factor_by(node) and min_depth <= current_depth:
            for child in children:
                for possible_child in _factorize(child):
                    split_node_with_child_assigned = assign_child(
                        node.copy(),  # type: ignore
                        possible_child,
                    )
                    yield split_node_with_child_assigned  # type: ignore

        case Node(nodes=children):
            # We need to return N copies of this node, with each
            # enumerating over all the posibilities of its children
            # e.g.
            # | children_sets = ((1, 2), (3, 4), (5,))
            # | for child_set in [(1, 3, 5,), (1, 4, 5,), (2, 3, 5,), (2, 4, 5,)]:
            # |    yield node.mutate(nodes=child_set)
            children_sets = (_factorize(c) for c in children)
            for child_set in product(*children_sets):
                yield node.mutate(nodes=child_set)



"""A value that is stored on disk and loaded lazily.

This is useful for transmitting large objects between processes.

```python exec="true" source="material-block" result="python" title="StoredValue"
from amltk.store import StoredValue
import pandas as pd
from pathlib import Path

df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
path = Path("df.csv")
df.to_csv(path)

stored_value = StoredValue(path, read=pd.read_csv)

# Somewhere in a processes
df = stored_value.value()
print(df)

path.unlink()
```

You can quickly obtain these from buckets if you require
```python exec="true" source="material-block" result="python" title="StoredValue from bucket"
from amltk import PathBucket
import pandas as pd

df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
bucket = PathBucket("bucket_path")
bucket.update({"df.csv": df})

stored_value = bucket["df.csv"].as_stored_value()

# Somewhere in a processes
df = stored_value.value()
print(df)

bucket.rmdir()
```
"""  # noqa: E501
from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Generic, TypeVar

K = TypeVar("K")
V = TypeVar("V")


@dataclass
class StoredValue(Generic[K, V]):
    """A value that is stored on disk and can be loaded when needed."""

    key: K
    read: Callable[[K], V]

    _value: V | None = None

    def value(self) -> V:
        """Get the value."""
        if self._value is None:
            self._value = self.read(self.key)

        return self._value



"""The
[`WarningFilter`][amltk.scheduling.plugins.warning_filter.WarningFilter]
if used to automatically filter out warnings from a [`Task`][amltk.scheduling.task.Task]
as it runs.

This wraps your function in context manager
[`warnings.catch_warnings()`][warnings.catch_warnings]
and applies your arguments to [`warnings.filterwarnings()`][warnings.filterwarnings],
as you would normally filter warnings in Python.

??? tip "Usage"

    ```python exec="true" source="material-block" html="true"
    import warnings
    from amltk.scheduling import Scheduler
    from amltk.scheduling.plugins import WarningFilter

    def f() -> None:
        warnings.warn("This is a warning")

    scheduler = Scheduler.with_processes(1)
    task = scheduler.task(f, plugins=WarningFilter("ignore"))
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
"""
from __future__ import annotations

import warnings
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeVar
from typing_extensions import ParamSpec, Self, override

from amltk.scheduling.plugins.plugin import Plugin

if TYPE_CHECKING:
    from rich.panel import Panel

    from amltk.scheduling.task import Task

P = ParamSpec("P")
R = TypeVar("R")
TrialInfo = TypeVar("TrialInfo")


class _IgnoreWarningWrapper(Generic[P, R]):
    """A wrapper to ignore warnings."""

    def __init__(
        self,
        fn: Callable[P, R],
        *warning_args: Any,
        **warning_kwargs: Any,
    ):
        """Initialize the wrapper.

        Args:
            fn: The function to wrap.
            *warning_args: arguments to pass to
                [`warnings.filterwarnings()`][warnings.filterwarnings].
            **warning_kwargs: keyword arguments to pass to
                [`warnings.filterwarnings()`][warnings.filterwarnings].
        """
        super().__init__()
        self.fn = fn
        self.warning_args = warning_args
        self.warning_kwargs = warning_kwargs

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        with warnings.catch_warnings():
            warnings.filterwarnings(*self.warning_args, **self.warning_kwargs)
            return self.fn(*args, **kwargs)


class WarningFilter(Plugin):
    """A plugin that disables warnings emitted from tasks."""

    name: ClassVar = "warning-filter"
    """The name of the plugin."""

    def __init__(self, *args: Any, **kwargs: Any):
        """Initialize the plugin.

        Args:
            *args: arguments to pass to
                [`warnings.filterwarnings`][warnings.filterwarnings].
            **kwargs: keyword arguments to pass to
                [`warnings.filterwarnings`][warnings.filterwarnings].
        """
        super().__init__()
        self.task: Task | None = None
        self.warning_args = args
        self.warning_kwargs = kwargs

    @override
    def attach_task(self, task: Task) -> None:
        """Attach the plugin to a task."""
        self.task = task

    @override
    def pre_submit(
        self,
        fn: Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> tuple[Callable[P, R], tuple, dict]:
        """Pre-submit hook.

        Wraps the function to ignore warnings.
        """
        wrapped_f = _IgnoreWarningWrapper(fn, *self.warning_args, **self.warning_kwargs)
        return wrapped_f, args, kwargs

    @override
    def copy(self) -> Self:
        """Return a copy of the plugin."""
        return self.__class__(*self.warning_args, **self.warning_kwargs)

    @override
    def __rich__(self) -> Panel:
        from rich.panel import Panel
        from rich.pretty import Pretty
        from rich.table import Table

        table = Table("Args", "Kwargs", padding=(0, 1), show_edge=False, box=None)
        table.add_row(Pretty(self.warning_args), Pretty(self.warning_kwargs))
        return Panel(table, title=f"Plugin {self.name}")



from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pandas as pd
from pytest_cases import case, parametrize, parametrize_with_cases

from amltk.distances import DistanceMetric, NearestNeighborsDistance, distance_metrics


@case
@parametrize("metric", distance_metrics.values())
def case_default_metric(metric: DistanceMetric) -> DistanceMetric:
    return metric


@case(tags="nn")
def case_nn_metric_l1() -> NearestNeighborsDistance:
    return NearestNeighborsDistance(algorithm="brute", metric="l1")


@case(tags="nn")
def case_nn_metric_l2() -> NearestNeighborsDistance:
    return NearestNeighborsDistance(algorithm="brute", metric="l2")


@case(tags="nn")
def case_nn_ball_tree() -> NearestNeighborsDistance:
    return NearestNeighborsDistance(algorithm="ball_tree")


@case(tags="nn")
def case_nn_auto() -> NearestNeighborsDistance:
    return NearestNeighborsDistance(algorithm="auto")


@parametrize_with_cases("metric", cases=".")
@parametrize(
    "target",
    [
        pd.Series([1, 2, 3], name="target", index=["mf1", "mf2", "mf3"]),
        np.array([1, 2, 3]),
        [1, 2, 3],
    ],
    idgen=lambda target: f"{target.__class__.__name__}",
)
@parametrize(
    "other",
    [
        pd.Series([1, 2, 3], name="other", index=["mf1", "mf2", "mf3"]),
        np.array([1, 2, 3]),
        [1, 2, 3],
    ],
    idgen=lambda other: f"{other.__class__.__name__}",
)
def test_distance_to_itself_is_zero(
    target: npt.ArrayLike,
    other: npt.ArrayLike,
    metric: DistanceMetric,
) -> None:
    assert metric(target, other) == 0



"""A mixin class for allowing rich output."""
from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import closing
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from rich.console import RenderableType


class RichRenderable(ABC):
    """Mixin for adding rich methods to a class."""

    def _repr_html_(self) -> str:
        """Return an HTML representation of the object."""
        return self._repr_pretty_()

    def _repr_pretty_(self, *_: Any, **__: Any) -> str:
        """Representation for rich printing."""
        from io import StringIO

        import rich

        with closing(StringIO()) as buffer:
            rich.print(self.__rich__(), file=buffer)
            return buffer.getvalue()

    @abstractmethod
    def __rich__(self) -> RenderableType:
        """Return a rich Text object."""



from amltk.sklearn.data import split_data, train_val_test_split
from amltk.sklearn.estimators import (
    StoredPredictionClassifier,
    StoredPredictionRegressor,
)
from amltk.sklearn.voting import voting_with_preffited_estimators

__all__ = [
    "train_val_test_split",
    "split_data",
    "StoredPredictionRegressor",
    "StoredPredictionClassifier",
    "voting_with_preffited_estimators",
]



"""A concurrent.futures.Executor interface that forces sequential execution."""
from __future__ import annotations

from collections.abc import Callable
from concurrent.futures import Executor, Future
from typing import TypeVar
from typing_extensions import ParamSpec, override

R = TypeVar("R")
P = ParamSpec("P")


class SequentialExecutor(Executor):
    """A [Executor][concurrent.futures.Executor] interface for sequential execution."""

    @override
    def submit(
        self,
        fn: Callable[P, R],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Future[R]:
        """Submit a function to be executed.

        Args:
            fn: The function to execute.
            *args: The positional arguments to pass to the function.
            **kwargs: The keyword arguments to pass to the function.

        Returns:
            A future that is already resolved with the result/exception of the function.
        """
        # TODO: It would be good if we can somehow wrap this in some sort
        # of async context such that it allows other callbacks to operate.
        future: Future[R] = Future()
        future.set_running_or_notify_cancel()

        try:
            future.set_result(fn(*args, **kwargs))
        except BaseException as exc:  # noqa: BLE001
            future.set_exception(exc)

        return future



from __future__ import annotations

from typing import TypeVar

import numpy as np
from pytest_cases import case, parametrize_with_cases

from amltk.randomness import as_int, as_randomstate, as_rng
from amltk.types import Seed

S = TypeVar("S", bound=Seed)


@case(tags=["static"])
def case_int_seed() -> tuple[int, int]:
    return 37, 37


@case(tags=["dynamic"])
def case_randomstate_seed() -> tuple[np.random.RandomState, np.random.RandomState]:
    return np.random.RandomState(137), np.random.RandomState(137)


@case(tags=["dynamic"])
def case_generator_seed() -> tuple[np.random.Generator, np.random.Generator]:
    return np.random.default_rng(1337), np.random.default_rng(1337)


@parametrize_with_cases("seeds", cases=".", has_tag="static")
def test_as_int_static(seeds: tuple[S, S]):
    seed1, seed2 = seeds

    first_i1, first_i2 = as_int(seed1), as_int(seed2)
    second_i1, second_i2 = as_int(seed1), as_int(seed2)

    # The first time we call it **should** be the same as the second time
    assert first_i1 == second_i1
    assert first_i2 == second_i2


@parametrize_with_cases("seeds", cases=".", has_tag="dynamic")
def test_as_int_dynamic(seeds: tuple[S, S]):
    seed1, seed2 = seeds

    first_i1, first_i2 = as_int(seed1), as_int(seed2)
    second_i1, second_i2 = as_int(seed1), as_int(seed2)

    assert first_i1 == first_i2
    assert second_i1 == second_i2

    # The first time we call it **should not** be the same as the second time
    assert first_i1 != second_i1
    assert first_i2 != second_i2


@parametrize_with_cases(argnames="seeds", cases=".", has_tag="static")
def test_as_randomstate_static(seeds: tuple[S, S]):
    seed1, seed2 = seeds

    first_i1, first_i2 = as_randomstate(seed1), as_randomstate(seed2)
    second_i1, second_i2 = as_randomstate(seed1), as_randomstate(seed2)

    first_ten_numbers_one = [first_i1.randint(0, 100) for _ in range(10)]
    first_ten_numbers_two = [first_i2.randint(0, 100) for _ in range(10)]

    assert first_ten_numbers_one == first_ten_numbers_two

    second_ten_numbers_one = [second_i1.randint(0, 100) for _ in range(10)]
    second_ten_numbers_two = [second_i2.randint(0, 100) for _ in range(10)]

    assert second_ten_numbers_one == second_ten_numbers_two

    assert first_ten_numbers_one == second_ten_numbers_one
    assert first_ten_numbers_two == second_ten_numbers_two


@parametrize_with_cases(argnames="seeds", cases=".", has_tag="dynamic")
def test_as_randomstate_dynamic(seeds: tuple[S, S]):
    seed1, seed2 = seeds

    first_i1, first_i2 = as_randomstate(seed1), as_randomstate(seed2)
    second_i1, second_i2 = as_randomstate(seed1), as_randomstate(seed2)

    first_ten_numbers_one = [first_i1.randint(0, 100) for _ in range(10)]
    first_ten_numbers_two = [first_i2.randint(0, 100) for _ in range(10)]

    assert first_ten_numbers_one == first_ten_numbers_two

    second_ten_numbers_one = [second_i1.randint(0, 100) for _ in range(10)]
    second_ten_numbers_two = [second_i2.randint(0, 100) for _ in range(10)]

    assert second_ten_numbers_one == second_ten_numbers_two

    assert first_ten_numbers_one != second_ten_numbers_one
    assert first_ten_numbers_two != second_ten_numbers_two


@parametrize_with_cases(argnames="seeds", cases=".", has_tag="static")
def test_as_rng_static(seeds: tuple[S, S]):
    seed1, seed2 = seeds

    first_i1, first_i2 = as_rng(seed1), as_rng(seed2)
    second_i1, second_i2 = as_rng(seed1), as_rng(seed2)

    first_ten_numbers_one = [first_i1.integers(0, 100) for _ in range(10)]
    first_ten_numbers_two = [first_i2.integers(0, 100) for _ in range(10)]

    assert first_ten_numbers_one == first_ten_numbers_two

    second_ten_numbers_one = [second_i1.integers(0, 100) for _ in range(10)]
    second_ten_numbers_two = [second_i2.integers(0, 100) for _ in range(10)]

    assert second_ten_numbers_one == second_ten_numbers_two

    assert first_ten_numbers_one == second_ten_numbers_one
    assert first_ten_numbers_two == second_ten_numbers_two


@parametrize_with_cases(argnames="seeds", cases=".", has_tag="dynamic")
def test_as_rng_dynamic(seeds: tuple[S, S]):
    seed1, seed2 = seeds

    first_i1, first_i2 = as_rng(seed1), as_rng(seed2)
    second_i1, second_i2 = as_rng(seed1), as_rng(seed2)

    first_ten_numbers_one = [first_i1.integers(0, 100) for _ in range(10)]
    first_ten_numbers_two = [first_i2.integers(0, 100) for _ in range(10)]

    assert first_ten_numbers_one == first_ten_numbers_two

    second_ten_numbers_one = [second_i1.integers(0, 100) for _ in range(10)]
    second_ten_numbers_two = [second_i2.integers(0, 100) for _ in range(10)]

    assert second_ten_numbers_one == second_ten_numbers_two

    assert first_ten_numbers_one != second_ten_numbers_one
    assert first_ten_numbers_two != second_ten_numbers_two
